"""RowDiff: Produce diff-similar results between table extensions of two
FITS files or HDU lists. This can be a command-line module or class in a
script. Written to add this functionality to crds.diff.
"""
import sys

import difflib
from itertools import product
import numpy as np

# ==========================================================================

# Deferred
# from astropy.table import Table
# from astropy.io.fits import TableDataDiff
# from astropy.io.fits.hdu.hdulist import fitsopen
# from astropy.io.fits.hdu.table import _TableLikeHDU

# ==========================================================================

from crds.core  import cmdline

# ==========================================================================
# Utilities


def table_to_string(a_table):
    """Convert a table to a string list.

    Parameters
    ----------
    a_table : astropy.table.table.Table
        The table to convert to a string


    Returns
    -------
    result : sequence of strings
        A sequence of strings, where each string is one row with comma-separated
        column values

    """
    result = list()
    for element in a_table:
        result.append(str(list(element)).strip('[]'))
    return result


def list_intersection(a_list, b_list, transform=lambda element: element):
    """Return a list of the intersection of two lists.

    Parameters
    ----------
    a_list : sequence
        First list. The order of the returned list is determined by this list.

    b_list : sequence or set
        Second list.


    Returns
    -------
    result : list
        The intersection of the input lists with the transformation performed
        on each element

    Other Parameters
    ----------------
    transform : function(element)
        A function to perform on each element while the intersection is
        being executed.

"""

    # Convert the second list to a set, if necessary
    b_set = b_list
    if not isinstance(b_set, set):
        b_set = set(b_list)

    # Do the intersection and return a list.
    result = [transform(value) for value in a_list
              if transform(value) in b_set]

    return result


def get_hdulist(fits_reference):
    """ Open an HDU list from the fits reference.
    Note that the reference may already be an HDUList
    """
    # If the reference is a string, presume its a file path
    from astropy.io.fits.hdu.hdulist import fitsopen
    result = fits_reference
    if isinstance(fits_reference, str):
        try:
            result = fitsopen(fits_reference)
        except Exception as exc:
            raise IOError("error opening file (%s): %s: %s" %
                          (fits_reference, exc.__class__.__name__, exc.args[0])) from exc
    return result


def hdus_consistent(a_hdulist, b_hdulist):
    """Check that two HDUs are consistent

    Parameters
    ----------
        a_hdulist, b_hdulist: pyfits.HDUList

    Returns
    -------
    result : bool
        Return True

    Raises
    ------
    RuntimeError
        Will be raised if any of the following occur:
            - Different number of extensions
            - Corresponding extensions are of different types.

    """

    # Check that there are the same number of HDUs
    if len(a_hdulist) != len(b_hdulist):
        raise RuntimeError('HDUs have different lengths: FITS A = %d, FITS B = %d' % \
            (len(a_hdulist), len(b_hdulist)))

    # Loop through the HDUs and check types
    # If one is different, then abort.
    for index in range(len(a_hdulist)):
        if a_hdulist[index].header.get('XTENSION') != \
           b_hdulist[index].header.get('XTENSION'):
            raise RuntimeError('Extension %d different: FITS A = %s, FITS B = %s' % \
                                   (index,
                                    a_hdulist[index].header.get('XTENSION'),
                                    b_hdulist[index].header.get('XTENSION')))

    return True

def column_name_lower(table):
    """Rename all columns to lowercase

    Parameters
    ----------
    table : astropy.table.Table
        Table to rename all columns for.

    Returns
    -------
    Modifies table in place

    """
    for name in table.colnames:
        try:
            table.rename_column(name, name.lower())
        except:
            pass


def report_mode_diff(diff):
    """Generate text for mode differences.

    Parameters
    ----------
    diff : tuple of (a_table, b_table, opcodes)
        `a_table` and b_table are the astropy.table.Table
        objects that difflib.SequenceMatcher
        was executed on. `opcodes` is that difference.

    Returns
    -------
    result :
        A single string containing a human-readable version of the difference

    """

    # Check edge case where no changes exist.
    result = ''
    if diff:
        (a_table,
         b_table,
         opcodes) = diff

        if len(opcodes):

            # Initialize the masks
            mask_deleted = np.zeros(len(a_table), dtype=bool)
            mask_added = np.zeros(len(b_table), dtype=bool)
            mask_changed_a = np.zeros(len(a_table), dtype=bool)
            mask_changed_b = np.zeros(len(b_table), dtype=bool)

            for opcode in opcodes:

                # Get the opcode info.
                (operation,
                 from_start,
                 from_end,
                 to_start,
                 to_end) = opcode

                # If a deletion, return eport row removed.
                if operation == 'delete':
                    mask_deleted[from_start:from_end] = True

                # If an addition, report rows added.
                elif operation == 'insert':
                    mask_added[to_start:to_end] = True

                # If changes, report the changes
                elif operation == 'replace':
                    mask_changed_a[from_start:from_end] = True
                    mask_changed_b[to_start:to_end] = True

            # Create the report tables.
            deleted = a_table[mask_deleted]
            added = b_table[mask_added]
            changed_a = a_table[mask_changed_a]
            changed_b = b_table[mask_changed_b]

            # Now report
            if len(deleted):
                result += '\n        Missing Modes:\n'
                result += str(deleted) + '\n'
            if len(added):
                result += '\n        Duplicated Modes:\n'
                result += str(added) + '\n'
            if len(changed_a):
                result += '\n        Changed Modes:\n'
                result += '        From Table A:\n'
                result += str(changed_a)
                result += '\n\n        To Table B:\n'
                result += str(changed_b) + '\n'

    # That's all folks
    return result


def sm_filter_opcodes(sm_opcodes, code='equal'):
    """Filter SequenceMatcher opcodes

    Parameters
    ----------
    sm_opcodes : sequence
        The result of difflib.SequenceMatcher.get_opcodes()

    code : string
        The code to remove.

    Returns
    -------
    result : The sequence with the specified operation removed.

    """

    # Define the filter function
    condition = lambda opcode: opcode[0] != code

    # do the filtering
    return list(filter(condition, sm_opcodes))


@np.vectorize
def selected(element, wanted):
    """Determine whether an element is in a sequence or set

    This is a conveninece routine to be decorated by numpy to
    allow efficient use on numpy collections

    Parameters
    ----------
    element : object
        Object to look for

    wanted : sequence or set
        The collection to look in

    Returns
    -------
    result : bool
        True if element is in the collection
    """

    return element in wanted


# ==========================================================================
class RowDiff:
    """Perform FITS table difference by rows

    Modules that are based on FITSDiff, such as Diff, compare
    tabular data on a column-by-column basis. Rowdiff compares tabular data
    on a row-by-row basis, producing UNIX diff-like output instead.
    Non-tabular extensions are ignored.

    Parameters
    ----------
    fits_a, fits_b: string or HDUList objects
        The two FITS files to compare. If strings are given, they
        are presumed to be file paths to open.

    fields: sequence
        List of fields to compare on.

    ignore_fields: sequence
        List of fields to ignore. Cannot specify both fields and ignore_fields

    mode_fields: sequence
        List of fields that define modes to compare


    Returns
    -------
    self.diffs : tuple for each table extension
        This is either None for no differences, or a
        tuple consisting of:
            - If mode_fields is specified, the tuple is described by
              modediff
            - Otherwise the tuple is described by rowdiff

    See Also
    --------
    self.rowdiff : Function that does the row differencing
    self.modediff : Function that does the mode differencing

    Notes
    -----
    The FITS data to be compared are required to be similar: they must have
    the same number of extensions and the types of extensions must match.

    The parameters --fields and --ignore-fields define which columns
    are compared between each table extension. These are mutually
    exclusive parameters and an error will generate if both are specified.

    """

    def __init__(self,
                 a_fits,
                 b_fits,
                 fields=[],
                 ignore_fields=[],
                 mode_fields=[]):

        self.a_hdulist = a_fits
        self.b_hdulist = b_fits
        self.fields = [field.lower() for field in fields]
        self.ignore_fields = [field.lower() for field in ignore_fields]
        self.mode_fields = mode_fields
        self.summary_only = False
        self.consistent = False

        # Check that fields and ignore_fields are not both
        # specified.
        if self.fields and self.ignore_fields:
            raise RuntimeError('Both fields and ignore_fields' +
                               ' cannot be specified.')

        # Get the FITS HDU's.
        with get_hdulist(a_fits) as self.a_hdulist, \
             get_hdulist(b_fits) as self.b_hdulist:

            # Do basic consistency checking. The number of HDUs in each
            # HDUList must be the same and the type of HDU must match.
            try:
                self.consistent = hdus_consistent(self.a_hdulist, self.b_hdulist)
            except RuntimeError as exc:
                print("rowdiff:", str(exc))

            # Set the differencing function.
            if self.mode_fields:
                self.diff = self.modediff
            else:
                self.diff = self.rowdiff

            # Loop through the HDU's, pick out the tables, then diff them.
            # We only need to look at the first HDUlist because we have
            # already checked consistency.
            self.diffs = []
            if not self.consistent:
                return

            from astropy.io.fits.hdu.table import _TableLikeHDU
            for hdu_index in range(len(self.a_hdulist)):
                if isinstance(self.a_hdulist[hdu_index], _TableLikeHDU):
                    self.diffs.append((hdu_index,
                                       self.diff(self.a_hdulist[hdu_index].data,
                                                 self.b_hdulist[hdu_index].data)))

    def modediff(self, a_fitstable, b_fitstable):
        """Produce diff-like output for table contents comparison

        Parameters
        ----------
        self: RowDiff instance
        a_fitstable: HDUList object
            First FITS table to compare
        b_fitstable: HDUList object
            Second FITS table to compare

        Returns
        -------
        result : tuple of diffs for the following comparisons:
            - All modes and table A
            - All modes and table B
            - table A vs. table B
            - Difference in rows of the common modes
              of both table A and table B
            Each item is either None for no differences,
            or a tuple consisting of
                - SequenceMatcher object
                - difflib.unified_diff object

        """
        # Get the mode fields.
        # self.mode_fields can also specify values for those fields.
        # If that is the case, we need to first separate out the
        # fields from the list
        if isinstance(self.mode_fields, dict):
            mode_field_names = dict.keys(self.mode_fields)

            mode_constraints = {key: value for key, value in self.mode_fields.items() if value is not None}
        else:
            mode_field_names = self.mode_fields
            mode_constraints = {}

        # Do a full FITS Table Data diff
        from astropy.io.fits import TableDataDiff
        data_diff = TableDataDiff(a_fitstable, b_fitstable,
                                  ignore_fields=self.ignore_fields,
                                  numdiffs=1)

        # Ensure that the mode select columns exist in both tables.
        # If all columns don't exit, then abort.
        if not set(mode_field_names) <= data_diff.common_column_names:
            raise RuntimeError('Mode select columns are not in both tables.')

        # Get the set of fields that will be compared.
        if self.fields:
            fields_common = list_intersection(self.fields,
                                              data_diff.common_column_names)
        else:
            # This funny business is done because common_column_names
            # is a set and the order is undetermined. This forces the
            # order to be the same as the original table.
            fields_common = list_intersection(a_fitstable.columns.names,
                                              data_diff.common_column_names,
                                              lambda element: element.lower())

        # Convert from FITS table to Astropy Table
        from astropy.table import Table    # Deferred
        a_table = Table(a_fitstable)
        b_table = Table(b_fitstable)

        # Rename the columns to lowercase.
        column_name_lower(a_table)
        column_name_lower(b_table)

        # Sort on the mode fields. We do this on the full tables
        # because later on we are going to examin the full tables
        # for diffing.
        a_table.sort(mode_field_names)
        b_table.sort(mode_field_names)

        # Create tables with just the mode columns
        a_table_modes = a_table[mode_field_names]
        b_table_modes = b_table[mode_field_names]

        values_possible = list()
        for field in mode_field_names:
            values_possible.append(sorted(list(set(list(a_table_modes[field]) +
                                                   list(b_table_modes[field])))))
        values_combinations = list(product(*values_possible))

        # Create a table of the combinations
        vc_array = np.array(values_combinations,
                            dtype=a_table_modes.dtype)
        vc_table = Table(vc_array)

        # Setup to do the diffing. Convert everything to strings.
        a_string = table_to_string(a_table_modes)
        b_string = table_to_string(b_table_modes)
        modes_string = table_to_string(vc_table)

        # Diff between the all modes and first table.
        sm = difflib.SequenceMatcher()
        sm.set_seqs(modes_string, a_string)
        result_modes_vs_a = (vc_table,
                             a_table_modes,
                             sm.get_opcodes())

        # Diff between all modes and the second table.
        sm.set_seqs(modes_string, b_string)
        result_modes_vs_b = (vc_table,
                             b_table_modes,
                             sm.get_opcodes())

        # Diff between the two tables.
        sm.set_seqs(a_string, b_string)
        result_a_vs_b = (a_table_modes,
                         b_table_modes,
                         sm.get_opcodes())
        matching_blocks = sm.get_matching_blocks()

        # For the last bit of magic, compare the common modes
        # between the two tables by rowdiff.
        # First see if there was any data difference,
        common_mode_diffs = None
        if data_diff.diff_rows or \
           data_diff.diff_values:

            # Rebuild the tables with just the matching modes and the
            # desired columns.
            if len(matching_blocks) > 1:

                # Create tables for just the fields of interest.
                # Need to include the mode columns also.
                # This is done against the original table to ensure
                # that the column order is the same.
                fields_all = list_intersection(a_table.colnames,
                                               mode_field_names +
                                               fields_common)

                a_table_common_fields = a_table[fields_all]
                b_table_common_fields = b_table[fields_all]

                # Create the tables that will be built into.
                dtypes = []
                for index in range(len(a_table_common_fields.dtype)):
                    dtypes.append(a_table_common_fields.dtype[index])
                a_table_values = Table(names=a_table_common_fields.colnames,
                                       dtype=dtypes)
                b_table_values = Table(names=a_table_common_fields.colnames,
                                       dtype=dtypes)

                for block in matching_blocks:
                    (a_row, b_row, n_rows) = block
                    for row in range(n_rows):
                        a_table_values.add_row(a_table_common_fields[a_row + row])
                        b_table_values.add_row(b_table_common_fields[b_row + row])

                # If values for the modes were defined, then select
                # only those rows.
                if isinstance(self.mode_fields, dict):
                    selected_rows = np.array([True for index in
                                              range(len(a_table_values))])
                    for (field, values) in dict.items(self.mode_fields):
                        if values:
                            selected_rows = np.logical_and(selected_rows,
                                                           selected(a_table_values[field], values))

                    a_table_values = a_table_values[selected_rows]
                    b_table_values = b_table_values[selected_rows]

                # We will be using string-based difflib for further
                # operations, so convert the tables to strings.
                a_string = table_to_string(a_table_values)
                b_string = table_to_string(b_table_values)

                # Find the differences. First use SequenceMatcher to get a
                # very concise list of changes.
                sm = difflib.SequenceMatcher()
                sm.set_seqs(a_string, b_string)
                common_mode_diffs = (a_table_values,
                                     b_table_values,
                                     sm.get_opcodes())

        # That's all folks.
        return (result_modes_vs_a,
                result_modes_vs_b,
                result_a_vs_b,
                common_mode_diffs)

    def rowdiff(self, a_fitstable, b_fitstable):
        """Produce diff-like output for table contents comparison

        Parameters
        ----------
        self: RowDiff instance
        a_fitstable: HDUList object
            First FITS table to compare
        b_fitstable: HDUList object
            Second FITS table to compare

        Returns
        -------
        result : Either None or a tuple of
            - SequenceMatcher object
            - String containing the unified diff version of the diff.

        """

        # Do a full FITS Table Data diff
        from astropy.io.fits import TableDataDiff
        data_diff = TableDataDiff(a_fitstable, b_fitstable,
                                  ignore_fields=self.ignore_fields,
                                  numdiffs=1)

        # Shortcut. If data differences, then end.
        if not data_diff.diff_rows and \
           not data_diff.diff_values:
            return None

        # Get the set of fields that will be compared.
        if self.fields:
            fields_common = list_intersection(self.fields,
                                              data_diff.common_column_names)
        else:
            # This funny business is done because common_column_names
            # is a set and the order is undetermined. This forces the
            # order to be the same as the original table.
            fields_common = list_intersection(a_fitstable.columns.names,
                                              data_diff.common_column_names,
                                              lambda element: element.lower())

        # If there are no common fields, exit
        if not fields_common:
            return None

        # Convert from FITS table to Astropy Table
        from astropy.table import Table
        a_table = Table(a_fitstable)
        b_table = Table(b_fitstable)

        # Rename the columns to lowercase.
        column_name_lower(a_table)
        column_name_lower(b_table)

        # Limit to the fields desired.
        a_table = a_table[fields_common]
        b_table = b_table[fields_common]

        # We will be using string-based difflib for further
        # operations, so convert the tables to strings.
        a_string = table_to_string(a_table)
        b_string = table_to_string(b_table)

        # Find the differences. First use SequenceMatcher to get a
        # very concise list of changes.
        sm = difflib.SequenceMatcher()
        sm.set_seqs(a_string, b_string)
        sm_opcodes = sm.get_opcodes()

        # If there are changes, produce a text diff of the changes.
        result = None
        if sm.ratio() < 0.99:
            unified_diff = difflib.unified_diff(a_string, b_string,
                                                "Table A", "Table B")
            result = (sm_opcodes, unified_diff)

        # Return the opcodes
        return result

    def __str__(self):
        """Provide readable output

        Returns
        -------
        result : String
            String of human readable version of the RowDiff object.

        """

        # Initialize the result string.
        result = ''

        # Save time, see if there are any diffs.
        if len(self.diffs) > 0:

            # Mode difference reporting
            if self.mode_fields:
                diff_current = 0
                for (diff_current, diff) in self.diffs:
                    result += 'Difference for HDU extension #%d\n' % diff_current

                    (result_modes_vs_a,
                     result_modes_vs_b,
                     result_a_vs_b,
                     common_mode_diffs) = diff

                    result_mode = report_mode_diff(result_modes_vs_a)
                    if result_mode:
                        result += '\n    Table A changes:\n'
                        result += result_mode
                    else:
                        result += '\n    Table A has all modes.\n'

                    result_mode = report_mode_diff(result_modes_vs_b)
                    if result_mode:
                        result += '\n    Table B changes:\n'
                        result += result_mode
                    else:
                        result += '\n    Table B has all modes.\n'

                    result_mode = report_mode_diff(result_a_vs_b)
                    if result_mode:
                        result += '\n    Table A to B changes:\n'
                        result += result_mode
                    else:
                        result += '\n    Table A and B share all modes.\n'

                    result_mode = report_mode_diff(common_mode_diffs)
                    if result_mode:
                        result += '\n    Common mode changes:\n'
                        result += '    If there were duplicate modes, the following may be nonsensical.\n'
                        result += result_mode
                    else:
                        result += '\n    All common modes are equivalent.\n'

            # Row difference reporting
            else:
                result_temp = ''
                for (diff_current, diff) in self.diffs:
                    if diff:
                        (sm_opcodes, unified_diff) = diff
                        filtered_diff = sm_filter_opcodes(sm_opcodes)
                        if filtered_diff:
                            result_temp += 'Row differences for HDU extension #%d\n\n' % diff_current
                            result_temp += '    Summary:\n'
                            for opcode in filtered_diff:

                                op, a_start, a_end, b_start, b_end = opcode
                                a_end -= 1
                                b_end -= 1

                                if op == 'replace':
                                    result_temp += '        a rows %d-%d differ from b rows %d-%d\n' % \
                                                   (a_start, a_end, b_start, b_end)
                                elif op == 'delete':
                                    result_temp += '        Remove from a rows %d-%d\n' % (a_start, a_end)
                                elif op == 'insert':
                                    result_temp += '        Add to b rows %d-%d\n' % (b_start, b_end)
                                else:
                                    pass

                            if not self.summary_only:
                                result_temp += '\n    Row difference, unified diff format:\n'
                                for unified_diff_row in unified_diff:
                                    result_temp += '        %s\n' % unified_diff_row

                if result_temp:
                    result += result_temp
                else:
                    result = '    HDU extension #%d contains no differences' % diff_current

        # Return the string
        return result

# =========================================================================
class RowDiffScript(cmdline.Script):
    """Command line script to perform FITS table difference by rows

    Parameters
    ----------
    fits_a, fits_b: Paths or HDUList objects of the
                    two FITS files to compare.
    fields: List of fields to compare on.
    ignore_fields: List of fields to ignore.
    mode_fields: List of fields that define modes to compare

    Note: The parameters 'fields' and 'ignore_fields' are mutually exclusive.
          An error will be raised if both are specified.

    Returns
    -------
      object variables:
          diffs: tuple of the differences for each table extension found.
                 This is either None for no differences, or is again a
                 tuple consisting of:
                     - If mode_fields is specified, the tuple is described by
                       modediff
                     - Otherwise the tuple is described by rowdiff
      stdout: Human readable report.

    """

    description = """Command line script to perform FITS table difference by rows"""

    epilog = """
    Modules that are based on FITSDiff, such as Diff, compare
    tabular data on a column-by-column basis. Rowdiff compares tabular data
    on a row-by-row basis, producing UNIX diff-like output instead.
    Non-tabular extensions are ignored.

    The FITS data to be compared are required to be similar: they must have
    the same number of extensions and the types of extensions must match.

    The parameters --fields and --ignore-fields define which columns
    are compared between each table extension. These are mutually
    exclusive parameters and an error will generate if both are specified.

    An example:

    % crds rowdiff s9m1329lu_off.fits s9518396u_off.fits
    Row differences for HDU extension #1

        Summary:
            Remove from a rows 0-3
            Add to b rows 4-35

        Row difference, unified diff format:
            --- Table A

            +++ Table B

            @@ -1,8 +1,36 @@

            -1, '1993-12-01', 224.84801, 2.5362999, -30.488701
            -2, '1993-12-01', 314.35199, -52.258701, -5.0489001
            -3, '1993-12-01', 44.669998, 0.87, 47.959999
            -4, '1993-12-01', 135.22099, 55.4618, -6.1592002
             1, '1994-01-25', 224.84801, 2.5362999, -30.488701

    First a summary of the changes between the table extension is given.
    Then, row-by-row difference is given, using unified diff syntax.

    The parameter --mode-fields initiates a different algorithm.
    Here, it is presumed the tabular data contains columns that can essentially
    be treated as keys upon with rows are selected. The fields specified are those
    key columns.

    All possible coombinations of values are determined be examining both
    extensions. Then, each table is compared against both this list and between
    each other, looking for multiply specified combinations, missing combinations,
    and, for the common combinations between the tables, whether the rest of the
    rows are equivalent or not.

    An example:
        % python -m rowdiff s9m1329lu_off.fits s9518396u_off.fits --mode-fields=detchip,obsdate
        Difference for HDU extension #1

            Table A changes:

                Missing Modes:
        detchip  obsdate
        ------- ----------
              1 1994-06-16
              1 1995-09-23
            <lines removed>

            Table B changes:

                Missing Modes:
        detchip  obsdate
        ------- ----------
              1 1993-12-01
              2 1993-12-01
            <lines removed>

            Table A to B changes:

                Missing Modes:
        detchip  obsdate
        ------- ----------
              1 1993-12-01

                Duplicated Modes:
        detchip  obsdate
        ------- ----------
              4 1994-06-16
              4 1995-09-23
            <lines removed>

                Changed Modes:
                From Table A:
        detchip  obsdate
        ------- ----------
              2 1993-12-01
              3 1993-12-01
              4 1993-12-01

                To Table B:
        detchip  obsdate
        ------- ----------
              1 1994-06-16
              1 1995-09-23
              1 1997-11-02
            <lines removed>

            All common modes are equivalent.

    """

    # Define the user interface
    def add_args(self):
        """Define the command line interface"""

        self.add_argument("tableA", help="First table to compare")
        self.add_argument("tableB", help="Second table to compare")
        self.add_argument("--ignore-fields", help="List of fields to ignore",
                          type=str)
        self.add_argument("--fields",
                          help="List of fields to compare",
                          type=str)
        self.add_argument("--mode-fields",
                          help="List of fields to do a mode compare",
                          type=str)

    locate_file = cmdline.Script.locate_file_outside_cache

    # Main program
    def main(self):
        """Perform the differencing"""

        # Get the path to the fits files.
        tableA_path = self.locate_file(self.args.tableA)
        tableB_path = self.locate_file(self.args.tableB)

        # Expand out the input field lists.
        fields = []
        ignore_fields = []
        mode_fields = []
        if self.args.fields is not None:
            fields = self.args.fields.split(',')
        if self.args.ignore_fields is not None:
            ignore_fields = self.args.ignore_fields.split(',')
        if self.args.mode_fields is not None:
            mode_fields = self.args.mode_fields.split(',')

        print(RowDiff(tableA_path, tableB_path,
                      fields=fields,
                      ignore_fields=ignore_fields,
                      mode_fields=mode_fields))

if __name__ == "__main__":
    sys.exit(RowDiffScript()())



"""This module defines replacement functionality for the CDBS "certify" program
used to check parameter values in .fits reference files.   It verifies that reference
files in multiple formats (FITS, json, yaml, ...) define required parameters and that
they have legal values.    It is also used to verify that CRDS mapping files are
consistent with outside systems.
"""
import os
from collections import defaultdict
import gc
import uuid
import re

import asdf
import numpy as np

# ============================================================================

import crds

from crds.core import pysh, log, config, utils, rmap, cmdline
from crds.core.exceptions import InvalidFormatError, ValidationError, MissingKeywordError, MappingInsertionError
from crds.core import reftypes

from crds import data_file, diff
from crds.io import tables
from crds.client import api
from crds.refactoring import refactor
# from crds.io import abstract

from . import mapping_parser
from . import validators
from . import check_sha1sum

# ============================================================================

class Certifier:
    """Baseclass for all certifiers: references, mappings, etc."""

    def __init__(self, filename, context, check_references=False,
                 compare_old_reference=False, dump_provenance=False,
                 provenance_keys=None,
                 dont_parse=False, script=None, observatory=None, comparison_reference=None,
                 original_name=None, run_fitsverify=False, check_sha1sum=False):

        self.filename = filename

        if context is None:
            raise ValueError("The 'context' argument must not be None")
        else:
            self.context = context

        self.check_references = check_references
        self.compare_old_reference = compare_old_reference
        self._dump_provenance_flag = dump_provenance
        self.dont_parse = dont_parse     # mapping only
        self.script = script
        self.comparison_reference = comparison_reference
        self.original_name = original_name
        self.run_fitsverify = run_fitsverify
        self.check_sha1sum = check_sha1sum
        self.error_on_exception = log.exception_trap_logger(self.log_and_track_error)

        assert self.check_references in [False, None, "exist", "contents"], \
            "invalid check_references parameter " + repr(self.check_references)

        self.observatory = observatory or utils.file_to_observatory(filename)

        self.provenance_keys = list(provenance_keys or utils.get_observatory_package(self.observatory).PROVENANCE_KEYWORDS)


    @property
    def basename(self):
        """Return the basename of the file being certified by this Certifier."""
        return os.path.basename(self.filename)

    @property
    def format_name(self):
        """Return the quoted name of the file being checked,  using the `original_name` substitute
        provided by web interfaces, or else the basename of the file or temporary file being certified.
        """
        return repr(self.original_name) if self.original_name else repr(self.basename)

    @property
    def locator(self):
        """Return the locator module for the observatory this Certifier corresponds to."""
        return utils.get_locator_module(self.observatory)

    def log_and_track_error(self, *args, **keys):
        """Output a log error on behalf of `msg`,  tracking it for uniqueness if run inside a script."""
        if self.script:
            self.script.log_and_track_error(self.filename, *args, **keys)
        else:
            log.error("In", repr(self.basename), ":", *args, **keys)

    def certify(self):
        """Certify `self.filename`,  either reporting using log.error() or raising
        exceptions.
        """
        raise NotImplementedError("Certify is an abstract class.")


    def get_validators(self):
        """Given a reference file `filename`,  return the observatory specific
        list of Validators used to check that reference file type.
        """
        # Get the cache key for this filetype.
        checkers = validators.get_validators(self.observatory, self.filename, context=self.context)
        checkers = self.set_rmap_parkeys_to_required(checkers)
        return checkers

    def set_rmap_parkeys_to_required(self, checkers):
        """Mutate copies of `checkers` so that any specified by the rmap parkey are required."""
        parkeys = set(self.get_rmap_parkeys())
        vlist = []
        for valid in checkers:
            if not valid.optional:
                vlist.append(valid)
            elif valid.name not in parkeys:
                vlist.append(valid)
            else:
                log.verbose("Mapping", repr(valid.name), "to REQUIRED based on rmap parkeys from",
                            repr(self.get_corresponding_rmap().basename))
                vlist.append(valid.get_required_copy())
        return vlist

    def get_corresponding_rmap(self):
        """Return the rmap which corresponds to self.filename under self.context."""
        pmap = crds.get_pickled_mapping(self.context, ignore_checksum="warn")  # reviewed
        instrument, filekind = pmap.locate.get_file_properties(self.filename)
        return pmap.get_imap(instrument).get_rmap(filekind)

    def get_rmap_parkeys(self):
        """Determine required parkeys in reference path `refname` according to pipeline
        mapping `context`.
        """
        try:
            return self.get_corresponding_rmap().get_required_parkeys()
        except Exception as exc:
            log.verbose_warning("Failed retrieving required parkeys:", str(exc))
            return []

# ============================================================================

class ReferenceCertifier(Certifier):
    """Baseclass for most reference file certifier classes.
    1. Check simple keywords against TPN files using the reftype's validators.
    2. Check mode tables against prior reference of comparison_context.
    3. Dump out keywords of interest.
    """
    def __init__(self, *args, **keys):
        super(ReferenceCertifier, self).__init__(*args, **keys)
        self.header = None
        self.validators = None
        self.all_simple_names = None
        self.mode_columns = None
        self.types = reftypes.get_types_object(self.observatory)

    def complex_init(self):
        """Can't do this until we at least know the file is loadable."""
        self.validators = self.get_validators()
        self.all_simple_names = [ val.name for val in self.validators if val.info.keytype == 'H' ]
        self.mode_columns = self.get_mode_column_names()

    def certify(self):
        """Certify `self.filename`,  either reporting using log.error() or raising
        ValidationError exceptions.
        """
        if self.check_sha1sum:
            with self.error_on_exception("Duplicate file check"):
                check_sha1sum.check_sha1sum(self.filename, observatory=self.observatory)

        self.complex_init()

        with self.error_on_exception("Error loading"):
            self.header = self.load()
        if not self.header:
            return

        display_header = dict(self.header)
        display_header.pop("__builtins__", None)
        log.verbose("Header:", log.PP(display_header), verbosity=55)

        for checker in self.validators:
            try:
                checker.check(self.filename, self.header)
                log.verbose("Checked", checker, verbosity=70)
            except Exception as exc:
                if not log.get_exception_trap():
                    raise
                presence = checker.is_applicable(self.header)
                if presence == "W":  # excludes "O"
                    log.warning("Checking", repr(checker.info.name), "failed:",
                                str(exc))
                else:
                    self.log_and_track_error(
                        "Checking", repr(checker.info.name),":", str(exc))

        # Table checks and provenance not associated with a single TpnInfo
        # NOTE: "W" doesn't downgrade colum checking exceptions to warning.
        with self.error_on_exception(
                "Checking reference modes for", repr(self.filename)):
            if self.mode_columns:
                self.certify_reference_modes()
        with self.error_on_exception(
                "Dumping provenance for", repr(self.filename)):
            if self._dump_provenance_flag:
                self.dump_provenance()

        with self.error_on_exception(
                "Checking ASDF Standard version for", repr(self.filename)):
                self.check_asdf_standard_version()

    def load(self):
        """Load and parse header from self.filename."""
        from crds.io import abstract
        # needed_keys=tuple([checker.complex_name for checker in self.validators])
        header = data_file.get_header(
            self.filename, (), self.original_name, self.observatory)
        header = self.map_ref_keys_to_dataset_keys(header)
        # header = self.cross_strap_instrument_keywords(header)
        header = self.add_array_keywords(header)
        header = data_file.convert_to_eval_header(header)
        header = abstract.ensure_keys_defined(header, needed_keys=[checker.complex_name for checker in self.validators])
        return header

    def map_ref_keys_to_dataset_keys(self, header):
        """Based on the rmap corresponding to this reference filename a`header`,  map keywords
        in `header` from the names used in reference files to the corresponding names matched in
        datasets.   Returns new `header`.
        """
        rmapping = None
        with log.verbose_warning_on_exception("No corresponding rmap"):
            rmapping = self.get_corresponding_rmap()
        if rmapping:
            with self.error_on_exception("Error mapping reference names and values to dataset names and values"):
                header = rmapping.locate.reference_keys_to_dataset_keys(rmapping, header)
        return header

    @property
    def array_validators(self):
        """Return the list of Validator objects that apply to arrays."""
        return [checker for checker in self.validators if checker.info.keytype in ["A","D"]]

    def add_array_keywords(self, header):
        """Add synthetic array keywords based on properties of the arrays mentioned in
        array validators to header.

        Mutates `header`.
        """
        header = dict(header)
        for checker in self.array_validators:
            # None is untried,  UNDEFINED is tried and failed.
            if header.get(checker.complex_name, None) == "UNDEFINED":
                continue
            # Load missing arrays,  or add data to loaded arrays from 'A' prior to 'D'.
            if ((checker.complex_name not in header) or
                (checker.info.keytype=="D" and header[checker.complex_name]["DATA"] is None)):
                header[checker.complex_name] = data_file.get_array_properties(self.filename, checker.name, checker.info.keytype)
        seen = set()
        for checker in self.array_validators:
            is_undefined = header.get(checker.complex_name, "UNDEFINED") == "UNDEFINED"
            if is_undefined:
                header[checker.complex_name] = "UNDEFINED"
                if checker.complex_name not in seen:
                    try:
                        checker.handle_missing(header)
                    except MissingKeywordError:
                        self.log_and_track_error("Missing required array", repr(checker.name))
                        seen.add(checker.complex_name)
        return header

    def dump_provenance(self):
        """Dump out provenance keywords for informational purposes."""
        dump_keys = sorted(
            set(key.upper() for key in
                self.get_rmap_parkeys() + # what's matched,  maybe not .tpn
                self.all_simple_names +   # what's defined in .tpn's, maybe not matched
                self.provenance_keys))    # extra project-specific keywords like HISTORY, COMMENT, PEDIGREE
        unseen = self._dump_provenance_core(dump_keys)
        log.verbose("Potential provenance keywords:", repr(dump_keys), verbosity=80)
        warn_keys = self.provenance_keys
        for key in sorted(unseen):
            if key in warn_keys:
                log.warning("Missing keyword '%s'."  % key)

    def _dump_provenance_core(self, dump_keys):
        """Generic dumper for self.header,  returns unseen keys."""
        unseen = set(dump_keys)
        for key in sorted(dump_keys):
            if self._check_provenance_key(key):
                unseen.remove(key)
        return unseen

    def _check_provenance_key(self, key):
        """Check one keyword, dump it,  and return True IFF it was present in self.header."""
        hval = self.header.get(key, None)
        if hval is not None:
            if self.interesting_value(hval):
                log.info(key, "=", repr(hval))
            return True
        return False

    def interesting_value(self, value):
        """Return True IFF `value` isn't uninteresting."""
        if str(value).strip().lower() in \
                ["",
                 "*** end of mandatory fields ***",
                 "*** column names ***",
                 "*** column formats ***"]:
            return False
        return True

    def get_mode_column_names(self):
        """Return any column names of `self` defined to be mode columns by the corresponding rmap in `self.context`.

        Only tables whose rmaps define row_keys will have mode checking performed.

        The first iteration of row_keys were defined as an rmap header paramter.  Subsequent iterations switched
        to a global definition in the locator module file rowkeys.dat.   The current iteration defines rowkeys in
        the spec for each type in the observatory package.
        """
        mode_columns = []
        with self.error_on_exception("Error finding unique row keys for", repr(self.basename)):
            instrument, filekind = utils.get_file_properties(self.observatory, self.filename)
            mode_columns = self.types.get_row_keys(instrument, filekind)
            if mode_columns:
                if tables.ntables(self.filename):
                    pass
                    # log.info("Potential table unique row selection parameters are", repr(mode_columns))
                    # log.info("Final combination is intersection with available table columns.")
                else:
                    log.verbose("No tables defined in reference.   Skipping row checks.")
            else:
                log.verbose("No unique row parameters, skipping table row checks.")
        return mode_columns

    def certify_reference_modes(self):
        """Check column parameters row-by-row, using mode groups."""
        if self.comparison_reference:
            old_reference = self.comparison_reference
        else:
            old_reference = self.find_old_reference(self.context, self.filename)
            if old_reference is None or old_reference == self.basename:
                # Load tables modes anyway,  looking for duplicate modes.
                for tab in tables.tables(self.filename):
                    table_mode_dictionary("new reference", tab, self.mode_columns)
                log.warning("No comparison reference for", repr(self.basename),
                            "in context", repr(self.context) + ". Skipping tables comparison.")
                return
        n_old_segments = tables.ntables(old_reference)
        n_new_segments = tables.ntables(self.filename)
        if n_old_segments != n_new_segments:
            log.warning("Differing HDU counts in", repr(old_reference), "and", repr(self.basename), ":",
                        n_old_segments, "vs.", n_new_segments)

        old_tables = tables.tables(old_reference)
        new_tables = tables.tables(self.filename)

        for i in range(0, min(n_new_segments, n_old_segments)):
            with self.error_on_exception("Checking tables modes in segment", i, "of", repr(self.filename)):
                self.check_table_modes(old_tables[i], new_tables[i])

    def find_old_reference(self, context, reffile):
        """Returns the name of the old reference file(s) that the new reffile would replace in `context`,  or None.
        """
        log.verbose("Resolving comparison reference for", repr(reffile), "in context", repr(context))
        with log.warn_on_exception("Failed resolving comparison reference for table checks"):
            return self._find_old_reference(context, reffile)

    def _find_old_reference(self, context, reffile):
        """Returns the name of the old reference file(s) that the new reffile would replace."""

        reference_mapping = find_governing_rmap(context, reffile)

        refname = os.path.basename(reffile)
        if refname in reference_mapping.reference_names():
            return refname

        # Determine the corresponding reference by attempting to add reffile to the old context.
        new_r = reference_mapping.insert_reference(reffile)

        # Examine the differences and treat the replaced file as the prior reference.
        diffs = reference_mapping.difference(new_r)
        match_refname = None
        for diff_tup in diffs:
            if diff.diff_action(diff_tup) == "replace":
                match_refname, dummy = diff.diff_replace_old_new(diff_tup)
                assert dummy == refname, "Bad replacement inserting '{}' into '{}'".format(reffile, reference_mapping.name)
                break
        else:
            log.info("No file corresponding to", repr(reffile), "in context", repr(reference_mapping.name))
            return None

        # grab match_file from server and copy it to a local disk, if network
        # connection is available and configured properly
        # Note: this call works in both networked and non-networked modes of operation.
        # Non-networked mode requires access to /grp/crds/[hst|jwst] or a copy of it.
        try:
            match_files = api.dump_references(reference_mapping.name, baserefs=[match_refname], ignore_cache=False)
            match_file = match_files[match_refname]
            if not os.path.exists(match_file):   # For server-less mode in debug environments w/o Central Store
                raise IOError("Comparison reference " + repr(match_refname) + " is defined but does not exist.")
            log.info("Comparing reference", repr(refname), "against", repr(os.path.basename(match_file)))
        except Exception as exc:
            log.warning("Failed to obtain reference comparison file", repr(match_refname), ":", str(exc))
            match_file = None

        return match_file

    def check_table_modes(self, old_table, new_table):
        """Check the tables modes of extension `ext` of `old_reference` versus self.filename"""
        new_reference_ex = new_table.basename + "[" + str(new_table.segment) + "]"
        old_reference_ex = old_table.basename + "[" + str(old_table.segment) + "]"
        log.verbose("Checking tables modes of '{}' against comparison reference '{}'".format(
            new_reference_ex, old_reference_ex))
        old_modes, old_all_cols = table_mode_dictionary("old reference", old_table, self.mode_columns)
        if not old_modes:
            log.info("No modes defined in comparison reference", repr(old_reference_ex),
                     "for keys", repr(self.mode_columns))
            return
        new_modes, new_all_cols = table_mode_dictionary("new reference", new_table, self.mode_columns)
        if not new_modes:
            log.info("No modes defined in new reference", repr(new_reference_ex), "for keys",
                     repr(self.mode_columns))
            return
        old_sample = list(old_modes.values())[0]
        new_sample = list(new_modes.values())[0]
        if len(old_sample) != len(new_sample) or old_all_cols != new_all_cols:
            log.warning("Change in row format between", repr(old_reference_ex), "and", repr(new_reference_ex))
            log.verbose("Old sample:", repr(old_sample))
            log.verbose("New sample:", repr(new_sample))
            return
        for mode in sorted(old_modes):
            if mode not in new_modes:
                log.warning("Table mode", mode, "from old reference", repr(old_reference_ex),
                            "is NOT IN new reference", repr(new_reference_ex))
                log.verbose("Old:", repr(old_modes[mode]), verbosity=60)
                continue
            # modes[mode][0] is row_no,  modes[mode][1] is row value
            diffs = self.compare_row_values(mode, old_modes[mode][1], new_modes[mode][1])
            if not diffs:
                log.verbose("Mode", mode, "of", repr(new_reference_ex),
                            "has same values as", repr(old_reference_ex),  verbosity=60)
            else:
                log.verbose("Mode change", mode, "between", repr(old_reference_ex), "and",
                            repr(new_reference_ex))
                log.verbose("Old:", repr(old_modes[mode]), verbosity=60)
                log.verbose("New:", repr(new_modes[mode]), verbosity=60)
        for mode in sorted(new_modes):
            if mode not in old_modes:
                log.info("Table mode", mode, "of new reference", repr(new_reference_ex),
                         "is NOT IN old reference", repr(old_table.basename))
                log.verbose("New:", repr(new_modes[mode]), verbosity=60)

    def compare_row_values(self, mode, old_row, new_row):
        """Compare key value tuple list `old_row` to `new_row` for key value tuple list `mode`.
        Handle array value comparisons.

        Return 0 if old_row == new_row,  non-0 otherwise.
        """
        different = 0
        for field_no, (old_key, old_value) in enumerate(old_row):
            new_key, new_value = new_row[field_no]
            if old_key != new_key:
                log.warning("Column key mismatch at mode", mode, "old_key", repr(old_key),
                            "new_key", new_key)
                different += 1
            old_value = handle_nan(old_value)
            new_value = handle_nan(new_value)
            if np.any(old_value != new_value):
                different += 1
        return different

    def check_asdf_standard_version(self):
        """
        If the file is an ASDF file or contains an embedded ASDF file,
        confirm that the file's ASDF Standard version obeys the context
        requirement.
        """
        asdf_standard_version = data_file.get_asdf_standard_version(self.filename)
        if asdf_standard_version:
            pmap = crds.get_pickled_mapping(self.context, ignore_checksum="warn")
            asdf_standard_requirement = pmap.get_asdf_standard_requirement()
            if not asdf_standard_version in asdf_standard_requirement:
                log.error(
                    "ASDF Standard version",
                    asdf_standard_version,
                    "does not fulfill context requirement of",
                    str(asdf_standard_requirement)
                )

# ============================================================================

def find_governing_rmap(context, reference):
    """Given mapping `context`,  return the loaded rmap which governs `reference`.   Typically this will
    be the rmap which contains the predecessor to `reference`,  not `reference` itself.
    """
    mapping = rmap.asmapping(context, cached=True)
    instrument, filekind = mapping.locate.get_file_properties(reference)
    if mapping.name.endswith(".pmap"):
        governing_rmap = mapping.get_imap(instrument).get_rmap(filekind)
    elif mapping.name.endswith(".imap"):
        governing_rmap = mapping.get_rmap(filekind)
    elif mapping.name.endswith(".rmap"):
        governing_rmap = mapping
    else:
        raise ValueError("Invalid comparison context " + repr(context))
    g_instrument, g_filekind = mapping.locate.get_file_properties(governing_rmap.name)
    assert instrument == g_instrument, "Comparison context inconsistent with reference file: " + repr(instrument) + " vs. " + repr(g_instrument)
    assert filekind == g_filekind, "Comparison context inconsistent with reference type: " + repr(filekind) + " vs. " + repr(g_filekind)
    log.verbose("Reference '{}' corresponds to rmap '{}' in context '{}'".format(
        reference, governing_rmap.name, mapping.name))
    return governing_rmap

# ============================================================================

def table_mode_dictionary(generic_name, tab, mode_keys):
    """Returns ({ (mode_val,...) : (row_no, (entire_row_values, ...)) },  [col_name, ...] )
    for crds.tables `tab` where column names `mode_keys` define the  columns to select for mode values.
    """
    all_cols = [name.upper() for name in tab.colnames]
    basename = repr(os.path.basename(tab.filename) + "[{}]".format(tab.segment))
    log.info("Mode columns defined by spec for", generic_name, basename, "are:", repr(mode_keys))
    log.info("All column names for this table", generic_name, basename, "are:", repr(all_cols))
    log.info("Checking for duplicate modes using intersection", sorted(list(set(mode_keys)&set(all_cols))))
    modes = defaultdict(list)
    for i, row in enumerate(tab.rows):
        new_row = tuple(zip(all_cols, (handle_nan(v) for v in row)))
        rowdict = dict(new_row)
        # Table row keys can vary by extension.  Have CRDS support a simple model of using
        # whichever mode_keys are present in a given row.
        mode = tuple((key, rowdict[key]) for key in mode_keys if key in rowdict)
        if not mode:
            log.info("Empty actual mode in", generic_name, basename, "with candidate mode columns", mode_keys)
            return {}, []
        modes[mode].append((i, new_row))
    for mode in sorted(modes.keys()):
        if len(modes[mode]) > 1:
            log.warning("Duplicate definitions in", generic_name, basename, "for mode:", mode, ":\n",
                        "\n".join([repr(row) for row in modes[mode]]))
    # modes[mode][0] is first instance of multiply defined mode.
    return { mode:modes[mode][0] for mode in modes }, all_cols

def handle_nan(var):
    """Map nan values to 'nan' so that 'nan' == 'nan'."""
    if isinstance(var, (np.float32, np.float64, np.float128)) and np.isnan(var):
        return 'nan'
    elif isinstance(var, np.ndarray) and var.shape == () and np.any(np.isnan(var)):
        return 'nan'
    else:
        return var

# ============================================================================

class FitsCertifier(ReferenceCertifier):
    """Certifier dedicated to FITS format references."""

    def __init__(self, *args, **keys):
        super(FitsCertifier, self).__init__(*args, **keys)
        if self.run_fitsverify:
            status, out = pysh.status_out_err("which fitsverify")
            if status == 0:
                log.verbose("fitsverify enabled and installled at", repr(out))
            else:
                log.warning("External fitsverify program is enabled but not found on PATH.")
                self.run_fitsverify = False

    def load(self):
        """Use pyfits to verify the FITS format of self.filename."""
        if not self.filename.endswith(".fits"):
            log.verbose("Skipping FITS verify for '%s'" % self.basename)
            return
        with data_file.fits_open_trapped(self.filename, checksum=bool(config.FITS_VERIFY_CHECKSUM)) as pfile:
            pfile.verify(option='exception') # validates all keywords
            if "ASDF" in pfile:
                self.log_and_track_error("FITS files must not include an ASDF extension")

        log.info("FITS file", repr(self.basename), "conforms to FITS standards.")
        return super(FitsCertifier, self).load()


    def _dump_provenance_core(self, dump_keys):
        """FITS provenance dumper,  works on multiple extensions.  Returns unseen keys."""
        with data_file.fits_open_trapped(self.filename) as hdulist:
            unseen = set(dump_keys)
            for i, hdu in enumerate(hdulist):
                for key in dump_keys:
                    for card in hdu.header.cards:
                        if card.keyword == key:
                            if self.interesting_value(card.value):
                                log.info("["+str(i)+"]", key, card.value, card.comment)
                            if key in unseen:
                                unseen.remove(key)
        unseen = super(FitsCertifier, self)._dump_provenance_core(unseen)
        return unseen

    def certify(self):
        """Run checks on FITS file."""
        super(FitsCertifier, self).certify()

        # Add-on fitsverify program from cfitsio authors
        if self.run_fitsverify:
            self.fitsverify()

        # Project-specific checks, for JWST instantiates data model.
        self.locator.project_check(self.filename, self.get_corresponding_rmap())

    def fitsverify(self):
        """Run optional external fitsverify program from cfitsio library, installed separately from CRDS."""
        log.info("Running fitsverify.")
        # subprocess stderr and stdout are combined into output
        filename = self.filename # quoted
        status, output = pysh.status_out_err("fitsverify ${filename}") # secure
        interpret_fitsverify_output(status, output)

# -------------------------------------------------------------------------------------------------

RECATEGORIZED_MESSAGE = {
    'Unregistered XTENSION value' : log.info,
    'The OGIP long string keyword convention is used' : log.info,
    'recommended LONGSTRN keyword.' : log.info,
    'checksum is not' : log.error,
    'Invalid CHECKSUM' : log.error,
    'dyld: Library not loaded:' : log.warning,  # bad fitsverify not running,  not a known problem with reference.
}

def interpret_fitsverify_output(status, output):
    """Re-issue captured fitsverify output as CRDS log messages,  elevating some cherry
    picked messages from WARNING to ERROR,  and likewise deemphasizing some fitsverify
    ERROR messages to CRDS WARNING messages.

    Fitsverify output is prefixed in CRDS log with >> to distinguish it as sub-program output.

    Integrating with CRDS log adds to ERROR and WARNING counters that ultimately pass/fail
    certified files and/or a reference file delivery.
    """
    errors, warnings, infos = log.status()
    for line in output.splitlines():
        if "Error:" in line or "Warning:" in line:
            for altered in RECATEGORIZED_MESSAGE:
                if altered in line:
                    RECATEGORIZED_MESSAGE[altered](">> RECATEGORIZED", line)
                    break
            else:
                func = log.error if "Error:" in line else log.warning
                func(">>", line)
        else:
            log.info(">>", line)
            infos += 1
    if status != 0:
        log.info("Fitsverify returned a NONZERO COMMAND LINE ERROR STATUS.")
        infos += 1   #  don't count status info below
    if log.warnings() - warnings:
        log.warning("Fitsverify output contains errors or warnings CRDS recategorizes as WARNINGs.")
    if log.errors() - errors:
        log.error("Fitsverify output contains errors or warnings CRDS recategorizes as ERRORs.")
    if log.infos() - infos:
        log.info("Fitsverify output contains errors or warnings CRDS recategorizes as INFOs.")

# ============================================================================

class UnknownCertifier(Certifier):
    """Certifier for unknown type,  currently a pass through with a warning."""

    def certify(self):
        """Certify an unknown format file."""
        log.warning("No certifier defined for", repr(self.basename))
        with log.augment_exception("Error parsing ", exception_class=InvalidFormatError):
            self.load()

    def load(self):
        """Load file of unknown type."""
        with open(self.filename, "rb") as handle:
            contents = handle.read()
        return contents

class AsdfCertifier(ReferenceCertifier):
    """Certifier for ADSF type,  invoke data models checks."""

    _VERSION_RE = re.compile(f"^[0-9]+\.[0-9]+\.[0-9]+$")

    def certify(self):
        """Certify an unknown format file."""
        super(AsdfCertifier, self).certify()

        with asdf.open(self.filename) as af:
            if not self._VERSION_RE.match(af["asdf_library"]["version"]):
                log.warning("File written with dev version of asdf library:", af["asdf_library"]["version"])

        self.check_schema()

        # Project-specific checks, for JWST instantiates data model.
        self.locator.project_check(self.filename, self.get_corresponding_rmap())

    def check_schema(self):
        """
        If schema_uri is set, verify that the file validates
        against that schema.
        """
        rmap = self.get_corresponding_rmap()
        if rmap.schema_uri is not None:
            # The file will be validated against the schema when it
            # is opened:
            with asdf.open(self.filename, custom_schema=rmap.schema_uri):
                pass

# ============================================================================

class MappingCertifier(Certifier):
    """Parameter container for certifying a mapping file,  and possibly it's references."""

    def certify(self):
        """Certify mapping `self.filename` relative to `self.context`."""
        if not self.dont_parse:
            parsing = mapping_parser.parse_mapping(self.filename)
            mapping_parser.check_duplicates(parsing)

        mapping = rmap.fetch_mapping(self.filename, ignore_checksum="warn")
        mapping.validate_mapping()

        # derived_from = mapping.get_derived_from()
        derived_from = find_old_mapping(self.context, self.filename)
        if derived_from is not None:
            if derived_from.name == self.basename:
                log.verbose("Mapping", repr(self.filename), "did not change relative to context", repr(self.context))
            else:
                if not self.basename.endswith(".pmap"):
                    log.info("Mapping", repr(self.basename), "corresponds to", repr(derived_from.name),
                             "from context", repr(self.context), "for checking mapping differences.")
                diff.mapping_check_diffs(mapping, derived_from)
        else:
            log.info("No predecessor for", repr(mapping.name), "relative to context", repr(self.context))

        # Optionally check nested references,  only for rmaps.
        if not self.check_references or not mapping.specifies_references:
            return

        references = self.get_existing_reference_paths(mapping)

        if self.check_references == "contents":
            certify_files(references, self.context,
                          dump_provenance=self._dump_provenance_flag,
                          check_references=self.check_references,
                          compare_old_reference=self.compare_old_reference,
                          script=self.script, observatory=self.observatory,
                          run_fitsverify=self.run_fitsverify,
                          check_rmap=False, check_sha1sums=False)

    def get_existing_reference_paths(self, mapping):
        """Return the paths of the references referred to by mapping.  Omit
        paths for which the reference does not exist.
        """
        references = []
        for ref in mapping.reference_names():
            path = None
            with self.error_on_exception("Can't locate reference file", repr(ref)):
                path = get_existing_path(ref, mapping.observatory)
            if path:
                log.verbose("Reference", repr(ref), "exists at", repr(path))
                references.append(path)
        return references

def get_existing_path(reference, observatory):
    """Return the path of `reference` located relative to `mapping`."""
    path = config.locate_file(reference, observatory)
    if not os.path.exists(path):
        raise ValidationError("Path " + repr(path) + " does not exist.")
    return path

def find_old_mapping(comparison_context, new_mapping):
    """Find the Mapping in pmap `comparison_context` corresponding to filename `new_mapping`,  if there is one.
    This call will cache `comparison_context` so it should only be called on "official" mappings,  not
    trial mappings.
    """
    if comparison_context:
        comparison_mapping = crds.get_pickled_mapping(comparison_context)  # reviewed
        old_mapping = comparison_mapping.get_equivalent_mapping(new_mapping)
        return old_mapping
    return None

def banner(char='#'):
    """Print a standard divider."""
    log.info(char * 40)  # Serves as demarkation for each file's report

# ============================================================================

def memory_cleanup(func):
    """Clear cached file data and collect garbage to prevent memory exhaustion."""
    def wrapped(*args, **keys):
        try:
            return func(*args, **keys)
        finally:
            data_file.clear_header_cache()
            tables.clear_cache()
            gc.collect()
    wrapped.__name__ = func.__name__ + "[memory_cleanup]"
    wrapped.__doc__ = func.__doc__
    return wrapped

# log.set_exception_trap short circuits deeply nested exception traps so that deep
# exceptions can propagate upward for debug.

@data_file.hijack_warnings
@memory_cleanup
def certify_file(filename, context, dump_provenance=False, check_references=False,
                 compare_old_reference=False,
                 dont_parse=False, script=None, observatory=None,
                 comparison_reference=None, original_name=None, ith="",
                 run_fitsverify=False, check_sha1sum=False):
    """Certify the list of `files` relative to .pmap `context`.   Files can be
    references or mappings.   This function primarily provides an interface for web code.

    filename:               path of file to certify
    context:                .pmap name to certify relative to
    dump_provenance:        for references,  log provenance keywords and rmap parkey values.
    check_references:       False, "exists", "contents"
    compare_old_reference:  bool,  if True,  attempt tables mode checking.
    dont_parse:             bool,  if True,  don't run parser to scan mappings for duplicate keys.
    script:                 command line Script instance
    original_name:          browser-side name of file if any, files
    """
    trap = log.error_on_exception if script is None else script.error_on_exception

    with trap(filename, "Certifier instantiation error"):
        if filename == "N/A":
            log.verbose("Skipping certify N/A file.")
            return

        original_name = filename if original_name is None else original_name
        observatory = utils.file_to_observatory(filename) if observatory is None else observatory

        filetype, klass = get_certifier_class(original_name, filename)

        if comparison_reference:
            log.info("Certifying", repr(original_name) + ith,  "as", repr(filetype.upper()),
                     "relative to context", repr(context), "and comparison reference", repr(comparison_reference))
        else:
            log.info("Certifying", repr(original_name) + ith, "as", repr(filetype.upper()),
                     "relative to context", repr(context))

        certifier = klass(filename, context, check_references=check_references,
                          compare_old_reference=compare_old_reference,
                          dump_provenance=dump_provenance,
                          dont_parse=dont_parse, script=script, observatory=observatory,
                          comparison_reference=comparison_reference,
                          original_name=original_name,
                          run_fitsverify=run_fitsverify,
                          check_sha1sum=check_sha1sum)

        with trap(filename, "Validation error"):
            certifier.certify()

def get_certifier_class(original_name, filepath):
    """Given a reference file name with a valid extension, return the filetype and
    Certifier subclass used to check it.
    """
    klasses = {
        "mapping" : MappingCertifier,
        "fits" : FitsCertifier,
        "json" : ReferenceCertifier,
        "yaml" : ReferenceCertifier,
        "asdf" : AsdfCertifier,
        "geis" : ReferenceCertifier,
        "unknown" : UnknownCertifier,
    }
    filetype = data_file.get_filetype(filepath, original_name)
    klass = klasses.get(filetype, UnknownCertifier)
    return filetype, klass

@memory_cleanup
def certify_files(files, context, dump_provenance=False, check_references=False,
                  compare_old_reference=False, dont_parse=False, skip_banner=False,
                  script=None, observatory=None, comparison_reference=None,
                  run_fitsverify=False, check_rmap=True, check_sha1sums=False):
    """Check the specified list of reference or mapping `files` paths.

    files:                  full paths of references or mappings to check
    context:                .pmap name to certify relative to
    dump_provenance:        for references,  log provenance keywords and rmap parkey values.
    check_references:       False, "exists", "contents"
    compare_old_reference:  bool,  if True,  attempt tables mode checking.
    dont_parse:             bool,  if True,  don't run parser to scan mappings for duplicate keys.
    skip_banner:            don't output the "#" separator lines between files
    script:                 command line Script instance
    observatory:            e.g. 'jwst' or 'hst'
    comparison_reference:   filepath to use for table comparison rather than finding in `context`.
    check_rmap:             run trial rmap update to check for overlapping reference cases.
    check_sha1sums:         check the sha1sums of `files` relative to files known on the CRDS server.
    """
    trap = log.error_on_exception if script is None else script.error_on_exception
    for fnum, filename in enumerate(files):

        if not skip_banner:
            banner()

        ith = ' (' + str(fnum+1) + '/' + str(len(files)) + ')'

        certify_file(
            filename, context, dump_provenance=dump_provenance, check_references=check_references,
            compare_old_reference=compare_old_reference, dont_parse=dont_parse, script=script, observatory=observatory,
            comparison_reference=comparison_reference, ith=ith, run_fitsverify=run_fitsverify, check_sha1sum=check_sha1sums)

    if check_rmap: # Requires checking all files in parallel, hence not in certify_file()
        if not skip_banner:
            banner()
        with trap("Failed updating rmap"):
            check_rmap_updates(observatory, context, files)

    if not skip_banner:
        banner()

# ============================================================================

@memory_cleanup
def check_rmap_updates(observatory, context, filepaths):
    """Do a test insertion of list of reference file paths `filepaths` into
    the appropriate rmaps under CRDS `context` for the purpose of detecting
    problems related to adding references to `context` as a group.

    observatory:  e.g. 'hst' or 'jwst'
    context: e.g. 'jwst_0499.pmap'
    filepaths:   [ reference_path, reference2_path, ...]

    The primary problem detected by the test insertion will be overlapping
    match cases which can happen between two of `filepaths` or
    between a new reference and an existing reference in the rmap.

    Overlaps come in two forms:

    1. In the extreme, a perfectly overlapping category will result in only
    one of two equivalent references being added to the rmap.

    2. When two categories overlap but one is a proper subset of the other.
    In this instance,  because the categories are different,  a replacement
    does not occur, but at runtime whenever dataset satisfying the more
    restrictive category occurs,  both categories match with equal weight;
    this results in an undesirable search ambiguity JWST disallows by default.
    """
    references = [ name for name in filepaths if config.is_reference(name) ]
    if not references:
        return
    observatory = utils.file_to_observatory(references[0]) if observatory is None else observatory
    organized = utils.organize_files(observatory, references)   # { (instrument, filekind) : [references,...] }
    pmap = crds.get_cached_mapping(context)
    for instrument, filekind in organized:
        references2 = organized[(instrument, filekind)]
        old_rmap = pmap.get_imap(instrument).get_rmap(filekind)
        new_rmap = "/tmp/" + old_rmap.basename
        log.info("Checking rmap update for", (instrument, filekind), "inserting files", references2)
        refactor.rmap_insert_references(old_rmap.filename, new_rmap, references2)

        banner()
        certify_file(new_rmap, context)    # check for partial overlaps

# ============================================================================

class CertifyScript(cmdline.Script, cmdline.UniqueErrorsMixin):
    """Command line script for checking CRDS mapping and reference files.

    Perform checks on each of `files`.   Print status.   If file is a context /
    mapping file,  it is used to define associated reference files which are
    located on the CRDS server.  If file is a .fits file,  it should include a
    relative or absolute filepath.
    """

    def __init__(self, *args, **keys):
#        super(CertifyScript, self).__init__(*args, **keys)
        if "print_status" not in keys:
            keys["print_status"] = True
        cmdline.Script.__init__(self, *args, **keys)
        cmdline.UniqueErrorsMixin.__init__(self, *args, **keys)

    description = """
Checks a CRDS reference or mapping file:

1. Verifies basic file format: .fits, .json, .yaml, .asdf, .pmap, .imap, .rmap
2. Checks references for required keywords and values, where constraints are defined.
3. Checks CRDS rules for permissible values with respect to defined reference constraints.
4. Checks CRDS rules for accidental file reversions or duplicate lines.
5. Checks CRDS rules for noteworthy version-to-version changes such as new or removed match cases.
6. Checks tables for deleted or duplicate rows relative to a comparison table.
7. Finds comparison references with respect to old CRDS contexts.
    """

    epilog = """

To run crds.certify on a reference(s) to verify basic file format and parameter constraints:

  % crds certify --comparison-context=hst_0027.pmap  ./some_reference.fits...

NOTE:  specifying ./ makes CRDS look in the current working directory instead of the CRDS cache.

If some_reference.fits is a table,  a comparison table will be found in the comparison context, if appropriate.

For recursively checking CRDS rules do this:

  % crds certify hst_0311.pmap --comparison-context=hst_0312.pmap

If a comparison context is defined, checked mappings will be compared against their peers (if they exist) in
the comparison context.  Many classes of mapping differences will result in warnings.

For reference table checks,  a comparison reference can also be specified directly rather than inferred from context:

  % crds certify ./some_reference.fits --comparison-reference=old_reference_version.fits

For more information on the checks being performed,  use --verbose or --verbosity=N where N > 50.
    """

    def add_args(self):
        self.add_argument("files", nargs="+")
        self.add_argument("-d", "--deep", dest="deep", action="store_true",
                          help="Certify reference files referred to by mappings have valid contents.")
        self.add_argument("-r", "--dont-recurse-mappings", dest="dont_recurse_mappings", action="store_true",
                          help="Do not load and validate mappings recursively,  checking only directly specified files.")
        self.add_argument("-a", "--dont-parse", dest="dont_parse", action="store_true",
                          help="Skip slow mapping parse based checks,  including mapping duplicate entry checking.")
        self.add_argument("-e", "--exist", dest="exist", action="store_true",
                          help="Certify reference files referred to by mappings exist.")
        self.add_argument("-p", "--dump-provenance", dest="dump_provenance", action="store_true",
                          help="Dump provenance keywords.")
        self.add_argument("-x", "--comparison-context", dest="comparison_context", type=str, default=None,
                          help="Pipeline context defining comparison files.  Defaults to operational context,  use 'none' to suppress.")
        self.add_argument("-y", "--comparison-reference", dest="comparison_reference", type=str, default=None,
                          help="Comparison reference for tables certification.")
        self.add_argument("-s", "--sync-files", dest="sync_files", action="store_true",
                          help="Fetch any missing files needed for the requested difference from the CRDS server.")
        self.add_argument("-l", "--allow-schema-violations", action="store_true",
                          help="Report jwst.datamodels schema violations as warnings rather than as errors.")
        self.add_argument("-f", "--run-fitsverify", action="store_true",
                          help="Run fitsverify for additional external checks on FITS files. cfitsio library must be installed separately.")
        self.add_argument("-u", "--check-rmap-updates", action="store_true",
                          help="Do a dry-run of adding reference files to the appropriate rmaps to detect errors.")
        self.add_argument("-k", "--check-sha1sums", action="store_true",
                          help="Check certified files to see if any are identical to files already in CRDS.")


        cmdline.UniqueErrorsMixin.add_args(self)

    # For files on the command line to default to normal UNIX syntax, no path
    # is CWD, uncomment following statement.  Add crds:// for cache paths.
    locate_file = cmdline.Script.locate_file_outside_cache

    def main(self):
        if self.args.deep:
            check_references = "contents"
        elif self.args.exist:
            check_references = "exist"
        else:
            check_references = None

        if self.args.allow_schema_violations:
            config.ALLOW_SCHEMA_VIOLATIONS.set(True)

        if not self.args.dont_recurse_mappings:
            all_files = self.mapping_closure(self.files)
        else:
            all_files = set(self.files)

        if self.args.comparison_context in ["none", "NONE", "None"]:
            log.warning("It is no longer possible to run the certifier without a comparison context.  Using default context instead.")
            self.args.comparison_context = None

        assert (self.args.comparison_context is None) or config.is_mapping_spec(self.args.comparison_context), \
            "Specified --context file " + repr(self.args.comparison_context) + " is not a CRDS mapping."
        assert (self.args.comparison_reference is None) or not config.is_mapping_spec(self.args.comparison_reference), \
            "Specified --comparison-reference file " + repr(self.args.comparison_reference) + " is not a reference."

        comparison_context = self._get_comparison_context(all_files)

        if self.args.comparison_reference:
            comparison_reference = config.locate_reference(self.args.comparison_reference, self.observatory)
        else:
            comparison_reference = None

        if self.args.sync_files:
            self._sync_comparison_files(comparison_context, comparison_reference)

        certify_files(sorted(all_files),
                      self.resolve_context(comparison_context),
                      comparison_reference=comparison_reference,
                      compare_old_reference=self.args.comparison_context or self.args.comparison_reference,
                      dump_provenance=self.args.dump_provenance,
                      check_references=check_references,
                      dont_parse=self.args.dont_parse,
                      script=self, observatory=self.observatory,
                      run_fitsverify=self.args.run_fitsverify,
                      check_rmap=self.args.check_rmap_updates,
                      check_sha1sums=self.args.check_sha1sums)

        self.dump_unique_errors()
        return log.errors()

    def _sync_comparison_files(self, comparison_context, comparison_reference):
        """Download comparison_context and comparison_reference as needed."""
        if comparison_context:
            resolved_context = self.resolve_context(comparison_context)
            self.sync_files([resolved_context])
        if comparison_reference:
            self.sync_files([comparison_reference])

    def _get_comparison_context(self, all_files):
        """Based on `all_files`,  --comparison-context, and --comparison-reference.

        Return any value for comparison_context (possibly defaulted to ops context) or None.
        """
        if self.args.comparison_context is None:  # no switch specified
            log.info("Defaulting --comparison-context to operational context.")
            comparison_context = self.default_context
        else:  # an explicit filename
            comparison_context = self.args.comparison_context
        return comparison_context

    def log_and_track_error(self, filename, *args, **keys):
        """Override log_and_track_error() to compute instrument, filekind automatically."""
        try:
            instrument, filekind = utils.get_file_properties(self.observatory, filename)
        except Exception:
            instrument = filekind = "unknown"
        super(CertifyScript, self).log_and_track_error(filename, instrument, filekind, *args, **keys)
        return None  # to suppress re-raise

    def mapping_closure(self, files):
        """Traverse the mappings in `files` and return a list of all mappings referred to by
        `files` as well as any references in `files`.
        """
        closure_files = set()
        for file_ in files:
            more_files = {file_}
            if config.is_mapping(file_):
                with self.error_on_exception(file_, "Problem loading submappings of", repr(file_)):
                    mapping = crds.get_cached_mapping(file_, ignore_checksum="warn")
                    more_files = {config.locate_mapping(name) for name in mapping.mapping_names()}
                    more_files = (more_files - {config.locate_mapping(mapping.basename)}) | {file_}
            closure_files |= more_files
        return sorted(closure_files)



"""This module defines functions for loading JWST's data model schema files which
describe reference parameters and their values.   The schema files are used to
validate reference and rmap parameters to screen out illegal values.

The primary functions provided by this module are:

1. Scraping the cal code data models schema for keyword value enumerations
2. Defining translations to and from "data model paths" like META.SUBARRAY.NAME and
"physical keyword values, e.g. FITS" like SUBARRAY.

See the tpn.py and locator.py modules,  as well as crds.certify and crds.rmap,
and crds.selectors for more information.
"""
import re

# ====================================================================================

# from jwst import datamodels   # deferred
# from . import locate  # deferred

import crds
from crds.core import log, utils, heavy_client, config
from crds.core.generic_tpn import TpnInfo

# ====================================================================================

INSTR_PREFIX = {
    "fgs" : "FGS_",
    "miri" : "MIR_",
    "nircam" : "NRC_",
    "niriss" : "NIS_",
    "nirspec" : "NRS_",
    }

def get_exptypes(instrument=None):
    """Using an arbitrary reference from an instrument that matches using EXP_TYPE,  return
    the set of all EXP_TYPE values defined in the JWST schema.   XXX kludged
    """
    tpninfos = get_schema_tpninfos(config.locate_file("jwst_nirspec_area_0005.fits", "jwst"))
    values = set()
    for info in tpninfos:
        if info.name in ["EXP_TYPE", "META.EXPOSURE.TYPE"]:
            values = values | set(info.values)
    if instrument is None:
        return sorted(list(values))
    else:
        return sorted([value for value in values
                       if value.startswith(INSTR_PREFIX[instrument.lower()])])

def get_schema_tpninfos(refpath):
    """Load the list of TPN info tuples corresponding to `instrument` and
    `filekind` from it's .tpn file.
    """
    with log.warn_on_exception("Failed loading schema constraints for", repr(refpath)):
        schema_name = reference_to_schema_name(refpath)
        tpns = get_schema_tpns(schema_name)
        parkeys = refpath_to_parkeys(refpath)
        return [ info for info in tpns if info.name in parkeys ]
    return []

def reference_to_schema_name(reference_name):
    """This function will eventually identify the schema associated with `reference_name`
    unless replaced by similar functionality in the models package.

    Returns None  meaning "default/core schema"
    """
    return None

@utils.cached
def get_schema_tpns(schema_name=None):
    """Load all the TpnInfos in the core schema."""
    flat = _schema_to_flat(_load_schema(schema_name))
    all_tpns = _flat_to_tpns(flat)
    return all_tpns

@utils.cached
def get_flat_schema(schema_name=None):
    """Flatten the specified data model schema, defaulting to the core schema,
    useful for retrieving FITS keywords or valid value lists.
    """
    return _schema_to_flat(_load_schema(schema_name))

@utils.cached
def refpath_to_parkeys(refpath):
    """Given a key for a TpnInfo's list, return the associated required parkeys."""
    from . import locate
    keys = []
    with log.verbose_warning_on_exception("Can't determine parkeys for", repr(refpath)):
        context  = heavy_client.get_context_name("jwst")
        p = crds.get_pickled_mapping(context)   # reviewed
        instrument, filekind = locate.get_file_properties(refpath)
        keys = p.get_imap(instrument).get_rmap(filekind).get_required_parkeys()
        # The below turned out to be bad because CRDS-only values aren't in datamodels
        # so in this case the datamodels schema are too narrow for these and will reject
        # e.g. SYSTEM or it's types.
        #         keys.append("META.INSTRUMENT.NAME")
        #         keys.append("META.REFTYPE")
    return sorted(keys)

# =============================================================================

# JWST Data Model schema to TPN


def _load_schema(schema_name=None):
    """Return the core data model schema."""
    from . import locate
    datamodels = locate.get_datamodels()
    model = datamodels.DataModel(schema=schema_name)
    return model.schema

def _schema_to_flat(schema):
    """Load the specified data model schema and return a flat dictionary from
    data model dotted path strings to TpnInfo objects.
    """
    flat = _x_schema_to_flat(schema)
    if flat is None:
        return None
    uppercase = {}
    for key, val in flat.items():
        if isinstance(val, list):
            val = [ str(item).upper() for item in val ]
        else:
            val = str(val).upper()
        uppercase[str(key).upper()] = val
    return uppercase

def _x_schema_to_flat(schema):
    """Recursively flatten `schema` without addressing case issues."""
    results = {}
    for feature in ["oneOf","allOf","$ref"]:
        if feature in schema:
            log.verbose_warning("Schema item has unhandled feature {}.", verbosity=80)
            return None

    if "anyOf" in schema and "type" in schema["anyOf"]:
        schema_type = schema["anyOf"]["type"]
    else:
        schema_type = schema.get("type", "null")

    if schema_type ==  "object":
        subprops = schema["properties"]
        for prop in subprops:
            with log.augment_exception("In schema property", repr(prop)):
                sub_tree = _schema_to_flat(subprops[prop])
                if sub_tree is None:
                    continue
                if isinstance(sub_tree, dict):
                    for subprop, val in list(sub_tree.items()):
                        results[prop + "." + subprop] = val
                else:
                    results[prop] = sub_tree
    elif schema_type in BASIC_TYPES:
        return schema
    elif schema_type in OPTIONAL_TYPES:
        return schema
    elif schema_type == "array":
        return None
    elif schema_type in ["any", "null"]:
        return None
    else:
        log.verbose_warning("Schema item has unhandled type", repr(schema_type), verbosity=80)
    return results

def type_or_null(names):
    """Return the list of types `names` + the name-or-null list for every type in `names`."""
    return [[name, 'null'] for name in names]

BASIC_TYPES = ["string","number","integer","boolean"]
OPTIONAL_TYPES = type_or_null(BASIC_TYPES)

#
# Only the first character of the field is stored, i.e. Header == H
#
# name = field identifier
# keytype = (Header|Group|Column)
# datatype = (Integer|Real|Logical|Double|Character)
# presence = (Optional|Required)
# values = [...]
#
# TpnInfo = namedtuple("TpnInfo", "name,keytype,datatype,presence,values")
#

SCHEMA_TYPE_TO_TPN = {
    "STRING" : ("C", "O"),
    "INTEGER" : ("I", "O"),
    "NUMBER" : ("D", "O"),
    "BOOLEAN" : ("L", "O"),

    ("STRING", "NULL") : ("C", "O"),
    ("INTEGER", "NULL") : ("I", "O"),
    ("NUMBER", "NULL") : ("D", "O"),
    ("BOOLEAN", "NULL") : ("L", "O"),
}

def _flat_to_tpns(flat=None, schema_name=None):
    """Convert flat representation of DM schema to list of all TpnInfo objects."""
    if flat is None:
        flat = _schema_to_flat(_load_schema(schema_name))
    tpns = []
    for key, value in flat.items():
        if key.endswith(".TYPE"):
            basekey = str(key[:-len(".TYPE")])
            legal_values = [str(val) for val in flat.get(basekey + ".ENUM", [])]
            if legal_values:
                legal_values += ["ANY", "N/A"]
            legal_values = tuple(sorted(set(legal_values)))
            if isinstance(value, list):
                value = tuple(value)
            datatype = SCHEMA_TYPE_TO_TPN.get(value, None)
            if datatype is not None:
                tpn = TpnInfo(name=basekey.upper(), keytype="H", datatype=datatype[0],
                              presence=datatype[1], values=legal_values)
                log.verbose("Adding tpn constraint from DM schema:", repr(tpn), verbosity=65)
                tpns.append(tpn)
            else:
                log.warning("No TPN form for", repr(key), repr(value))
    return sorted(tpns)

def _get_dm_to_fits(schema=None):
    """Return mapping from DM dotted path string to FITS keyword."""
    if schema is None:
        schema = _load_schema()
    fits = {}
    flat = _schema_to_flat(schema)
    for key, val in flat.items():
        if key.lower().endswith(".fits_keyword"):
            fits[str(key[:-len(".fits_keyword")])] = str(val)
    return fits

def _get_fits_to_dm(schema=None):
    """Return mapping from FITS keyword to DM dotted path string."""
    return utils.invert_dict(_get_dm_to_fits(schema))

DM_TO_FITS = None
FITS_TO_DM = None

def dm_to_fits(key):
    """Return the FITS keyword for DM `key` or None.

    >>> dm_to_fits('META.SUBARRAY.NAME')
    'SUBARRAY'
    """
    global DM_TO_FITS
    if DM_TO_FITS is None:
        DM_TO_FITS = _get_dm_to_fits()
    return DM_TO_FITS.get(key.upper(), None)

def fits_to_dm(key):
    """Return the DM keyword for FITS `key` or None.

    >>> fits_to_dm('SUBARRAY')
    'META.SUBARRAY.NAME'
    """
    global FITS_TO_DM
    if FITS_TO_DM is None:
        FITS_TO_DM = _get_fits_to_dm()
    return FITS_TO_DM.get(key.upper(), None)

# =============================================================================

def main():
    print("null tpn processing.")

if __name__ == "__main__":
    main()



'''
Created on Feb 15, 2017

@author: jmiller
'''
# import yaml

# ============================================================================

from crds.core import exceptions

from .abstract import AbstractFile
# ============================================================================

class YamlFile(AbstractFile):

    format = "YAML"

    def get_raw_header(self, needed_keys=(), **keys):
        """Return the flattened header associated with a YAML file."""
        import yaml
        with open(self.filepath) as pfile:
            try:
                header = yaml.safe_load(pfile)
            except ValueError as exc:
                raise exceptions.YamlFormatError(
                    "YAML wouldn't load from", repr(self.filepath), ":", str(exc))
            header = self.to_simple_types(header)
        return header



"""This module uses the package Parsley to parse CRDS mappings into a
header dictionary and a Selector Parameters object tree.  Selector
Parameters are uninstantiated Selectors which are first built up from
the mapping file.  Since the parser captures full item lists, with no
removal of duplicate keys, the results can be used to detect duplicate
dictionary entries which normally collide silently eliminating one
item.  This is principally intended to detect rmap cut-and-paste
errors in hand edited rmaps.
"""
from collections import namedtuple

from crds.core import rmap, selectors, log, exceptions, config

# NOTE:  #-comments are treated as white space and currently dropped when an rmap is rewritten
# as a new version.

MAPPING_GRAMMAR = r"""

ws = (' ' | '\r' | '\n' | '\t' | pound_comment)*

pound_comment = '#' (~'\n' anything)*:c '\n' -> ''.join(c)

mapping = header_section:h comment_section:c selector_section:s -> (h, s, c)

header_section = ws 'header' ws '=' ws dict:d ws -> d

comment_section = ((ws 'comment' ws '=' ws block_string:c ws -> c)
                  | -> None)
selector_section = ws 'selector' ws '=' ws selector:s ws -> s

selector = ( dict | parameters ):d -> d
parameters = ws parameters_name:n ws '(' ws dict:d ws ')' -> eval(n, globals(), locals())(d)
parameters_name = ('Match' | 'UseAfter' | 'VersionAfter' | 'SelectVersion' | 'Bracket' | 'GeometricallyNearest' | 'ClosestTime'):n -> n

# NOTE:  dict is returned as an item-list to preserve duplicates (considered errors)
dict = ws '{' items:m '}' ws -> m
items = ws (pair:first (ws ',' pair)*:rest ws ','? ws -> [first] + rest) | ws -> []
pair = ws immutable:k ws ':' ws value:v ws -> (k, v)
immutable = (string | immutable_tuple | number | true | false | none):i -> i

value = (string | tuple | number | selector | dict | set | true | false | none):v -> v

string = (('"' (escapedChar | ~'"' anything)*:c '"')
         |("'" (escapedChar | ~"'" anything)*:c "'") -> ''.join(c))

block_string = (('\"\"\"' (escapedChar | ~'\"\"\"' anything)*:c '\"\"\"')
               |("\'\'\'" (escapedChar | ~"\'\'\'" anything)*:c "\'\'\'")) -> ''.join(c)

escapedChar = '\\' (('"' -> '"')    |('\\' -> '\\')
                   |('/' -> '/')    |('b' -> '\b')
                   |('f' -> '\f')   |('n' -> '\n')
                   |('r' -> '\r')   |('t' -> '\t')
                   |('\'' -> '\'')  )

number = ('-' | -> ''):sign (intPart:ds (floatPart(sign ds) | -> int(sign + ds)))
digit = :x ?(x in '0123456789') -> x
digits = <digit*>
digit1_9 = :x ?(x in '123456789') -> x
intPart = (digit1_9:first digits:rest -> first + rest) | digit
floatPart :sign :ds = <('.' digits exponent?) | exponent>:tail -> float(sign + ds + tail)
exponent = ('e' | 'E') ('+' | '-')? digits

# list = '[' elements:xs ws ','? ws ']' -> list(xs)
tuple = '(' elements:xs ws ','? ws ')' -> tuple(xs)
set = '{' elements:xs ws ','? ws '}' -> set(xs)
elements = (ws value:first (ws ',' ws value)*:rest ws -> [first] + rest) | -> []

immutable_tuple = '(' immutable_elements:xs ws ','? ws ')' -> tuple(xs)
immutable_elements = (ws immutable:first (ws ',' ws immutable)*:rest ws -> [first] + rest) | -> []

true = 'True' -> True
false = 'False' -> False
none = 'None' -> None
"""

try:
    import parsley
except ImportError:
    parsley = None

MAPPING_PARSER = None

'''
def profile_parse(filename="hst_cos_deadtab.rmap"):
    """Profile the parsing of `filename`, print stats, and instantiate the
    mapping to run the duplicates checking.
    """
    import cProfile
    import pstats
    filename = config.locate_mapping(filename)
    statsname = os.path.splitext(filename)[0] + ".stats"
    cProfile.runctx('result = MAPPING_PARSER(open("{}").read()).mapping()'.format(filename),
                    locals(), globals(), statsname)
    stats = pstats.Stats(statsname)
    stats.sort_stats("time")
    stats.print_stats(20)
'''

Parsing = namedtuple("Parsing", "header,selector,comment")

def parse_mapping(filename):
    """Parse mapping `filename`.   Return parsing."""
    global parsley, MAPPING_PARSER

    if parsley is None:
        raise NotImplementedError("Parsley parsing package must be installed.")

    if MAPPING_PARSER is None:
        MAPPING_PARSER = parsley.makeGrammar(MAPPING_GRAMMAR, selectors.SELECTORS)

    log.verbose("Parsing", repr(filename))
    filename = config.locate_mapping(filename)
    with log.augment_exception("Parsing error in", repr(filename), exception_class=exceptions.MappingFormatError):
        with open(filename) as pfile:
            parser = MAPPING_PARSER(pfile.read())
            header, selector, comment = parser.mapping()
            return Parsing(header, selector, comment)

def check_duplicates(parsing):
    """Examine mapping `parsing` from parse_mapping() for duplicate header or selector entries."""
    if isinstance(parsing.selector, selectors.Parameters):
        parsing.selector.instantiate(parsing.header)
    else:
        selectors.check_duplicates(parsing.header, ["header"])
        selectors.check_duplicates(parsing.selector, ["selector"])



from .integration_tests import SynphotCoreIntegrationTest, SynphotObsmodeIntegrationTest
from .lookup_generator import SynphotLookupGenerator

__all__ = [
    "SynphotCoreIntegrationTest",
    "SynphotObsmodeIntegrationTest",
    "SynphotLookupGenerator",
]



"""This module provides functions which determine various observatory
specific policies/plugins for Roman:

1. How to convert reference file basenames to fully specified paths.

2. How to manage parameters for reference file Validator objects used
in the certification of reference files.

XXXX Roman NOTE: This code was derived from the JWST locate.py module and
contains substantial duplication.  However, because the functions often depend
on project-specific modules, globals, or functions the code is not usable
without some refactoring.  Other plugins may vary simply because
ASDF+datamodels Roman is already different than the FITS+datamodels world of
JWST, e.g. there is no longer a need for FITS <-> datamodels translations and
log annotation, i.e.  AKA keyword cross-strapping.

"""
import os.path
import re
import warnings
from collections import namedtuple

from asdf.tags.core import NDArrayType

# =======================================================================

from crds.core import rmap, config, utils, timestamp, log, exceptions
from crds.certify import generic_tpn
from crds import data_file
from crds.io import abstract

# =======================================================================

# These two functions decouple the generic reference file certifier program
# from observatory-unique ways of specifying and caching Validator parameters.

from crds.roman import TYPES, INSTRUMENTS, FILEKINDS, EXTENSIONS, INSTRUMENT_FIXERS, TYPE_FIXERS

get_row_keys_by_instrument = TYPES.get_row_keys_by_instrument
get_item = TYPES.get_item
suffix_to_filekind = TYPES.suffix_to_filekind
filekind_to_suffix = TYPES.filekind_to_suffix
get_all_tpninfos = TYPES.get_all_tpninfos

HERE = os.path.dirname(__file__) or "."

# =======================================================================

# XXXX roman  TODO needed for scratch JWST repro scheme,  itself incomplete
# from crds.jwst.pipeline import header_to_reftypes, header_to_pipelines

# Stub like HST for now

def header_to_reftypes(header, context="roman-operational"):
    """Based on `header` return the default list of appropriate reference type names.

    >>> ref_types = header_to_reftypes(None)
    >>> print(ref_types)
    []

    """
    return []  # translates to "all types" for instrument defined by header.

def header_to_pipelines(header, context="roman-operational"):
    """Based on `header` return the default list of appropriate reference type names.

    >>> header_to_pipelines(None)
    Traceback (most recent call last):
    ...
    NotImplementedError: Roman has not defined header_to_pipelines().

    """
    raise NotImplementedError("Roman has not defined header_to_pipelines().")

# =============================================================================

def tpn_path(tpn_file):
    """Return the full filepath of `tpn_file`.

    >>> tpn_path('tpn_file.tpn') # doctest: +ELLIPSIS
    '.../crds/roman/tpns/tpn_file.tpn'

    """
    return os.path.join(HERE, "tpns", tpn_file)

def get_extra_tpninfos(refpath):
    """Returns TpnInfos (valid value enumerations) derived from the cal code data models schema.

    This can be useful to leverage datamodels (core?) schema for rmap checking.

    Datamodels schema historically lack per-instrument granularity so the canonical .tpn scheme
    can catch errors not caught by the schema,  e.g. MIRI values used in a NIRCAM reference
    file because a reference file was cloned and not properly updated.

    These tend to have a superset of acceptable values for any particular instrument,
    particularly since CRDS has historically only loaded the core schema.

    >>> get_extra_tpninfos(None)
    []

    """
    return []   # use this to ignore datamodels schema for CRDS value checking
    # XXXX roman  -- datamodels schema scraping can potentially be enabled once
    # romancal and datamodels are integrated.   This will effectively translate the
    # datamodels schema (core unless work is done) automatically as-if they were
    # being specified in CRDS .tpn files.
    # return schema.get_schema_tpninfos(refpath)

def project_check(refpath, rmap):
    pass

# =======================================================================

def match_context_key(key):
    """Set the case of a context key appropriately for this project, Roman
    always uses upper case.

    >>> match_context_key('aB.$QqZ4nB')
    'AB.$QQZ4NB'

    """
    return key.upper()

# =======================================================================

@utils.cached
def get_file_properties(filename):
    """Figure out (instrument, filekind, serial) based on `filename` which
    should be a mapping or ASDF reference file.

    >>> get_file_properties('tests/data/roman_wfi16_f158_flat_small.asdf')
    ('wfi', 'flat')

    >>> get_file_properties('tests/data/roman_wfi_flat_0004.rmap')
    ('wfi', 'flat')

    >>> get_file_properties('tests/data/roman_0001.pmap')
    ('', '')

    >>> get_file_properties('tests/data/ascii_tab.csv') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: string indices must be integers

    """
    if config.is_mapping(filename):
        try:
            return decompose_newstyle_name(filename)[2:4]
        except Exception:
            # NOTE: load_mapping more conservative than fetch_mapping used in properties_from_mapping
            mapping = rmap.load_mapping(filename)
            return mapping.instrument, mapping.filekind
    elif config.is_reference(filename):
        result = get_reference_properties(filename)[2:4]
    else:
        try:
            result = properties_inside_mapping(filename)
        except Exception as exc:
            result = get_reference_properties(filename)[2:4]
    assert result[0] in INSTRUMENTS+[""], "Bad instrument " + \
        repr(result[0]) + " in filename " + repr(filename)
    assert result[1] in FILEKINDS+[""], "Bad filekind " + \
        repr(result[1]) + " in filename " + repr(filename)
    return result

def decompose_newstyle_name(filename):
    """
    >>> decompose_newstyle_name('./roman.pmap')
    ('.', 'roman', '', '', '', '.pmap')

    >>> decompose_newstyle_name('./roman_0001.pmap')
    ('.', 'roman', '', '', '0001', '.pmap')

    >>> decompose_newstyle_name("./roman_wfi_0001.imap")
    ('.', 'roman', 'wfi', '', '0001', '.imap')

    >>> decompose_newstyle_name("./roman_wfi_flat.rmap")
    ('.', 'roman', 'wfi', 'flat', '', '.rmap')

    >>> decompose_newstyle_name("./roman_wfi_flat.asdf")
    ('.', 'roman', 'wfi', 'flat', '', '.asdf')

    >>> decompose_newstyle_name("./hst_acs.imap")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid instrument 'acs'

    >>> decompose_newstyle_name("./roman_wfi_lark_0001.rmap")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid filekind 'dark'

    >>> decompose_newstyle_name("./roman_wfi_flat_abcd.rmap")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid id field <built-in function id>

    """
    path, parts, ext = _get_fields(filename)
    observatory = parts[0]
    serial = list_get(parts, 3, "")

    if ext == ".pmap":
        assert len(parts) in [1,2], "Invalid .pmap filename " + repr(filename)
        instrument, filekind = "", ""
        serial = list_get(parts, 1, "")
    elif ext == ".imap":
        assert len(parts) in [2,3], "Invalid .imap filename " + repr(filename)
        instrument = parts[1]
        filekind = ""
        serial = list_get(parts, 2, "")
    else:
        assert len(parts) in [3,4], "Invalid filename " + repr(filename)
        instrument = parts[1]
        filekind = parts[2]
        serial = list_get(parts, 3, "")

    # Don't include filename in these or it messes up crds.certify unique error tracking.

    assert instrument in INSTRUMENTS+[""], "Invalid instrument " + repr(instrument)
    assert filekind in FILEKINDS+[""], "Invalid filekind " + repr(filekind)
    assert re.fullmatch(r"\d*", serial), "Invalid id field " + repr(id)
    # extension may vary for upload temporary files.

    return path, observatory, instrument, filekind, serial, ext

def properties_inside_mapping(filename):
    """Load `filename`s mapping header to discover and
    return (instrument, filekind).

    >>> properties_inside_mapping('tests/data/roman_0001.pmap')
    ('', '')

    >>> properties_inside_mapping('tests/data/roman_wfi_flat_0004.rmap')
    ('wfi', 'flat')

    >>> properties_inside_mapping('tests/data/roman_wfi_0001.imap')
    ('wfi', '')

    >>> properties_inside_mapping('tests/data/roman_wfi_flat_0004.rmap')
    ('wfi', 'flat')

    """
    map = rmap.fetch_mapping(filename)
    if map.mapping == "pipeline":
        result = "", ""
    elif map.mapping == "instrument":
        result = map.instrument, ""
    else:
        result = map.instrument, map.filekind
    return result

def _get_fields(filename):
    """
    >>> _get_fields("")
    ('', [''], '')

    >>> _get_fields("a/b.c")
    ('a', ['b'], '.c')

    >>> _get_fields("_")
    ('', ['', ''], '')

    >>> _get_fields("__")
    ('', ['', '', ''], '')

    >>> _get_fields("a_b_c")
    ('', ['a', 'b', 'c'], '')

    >>> _get_fields("_a_b_c_")
    ('', ['', 'a', 'b', 'c', ''], '')

    """
    path = os.path.dirname(filename)
    name = os.path.basename(filename)
    name, ext = os.path.splitext(name)
    parts = name.split("_")
    return path, parts, ext

def list_get(l, index, default):
    """
    >>> list_get([], 0, None)

    >>> list_get([], -1, 7)
    7

    >>> list_get(None, 0, None)
    Traceback (most recent call last):
    ...
    TypeError: 'NoneType' object is not subscriptable

    >>> list_get([1], 1, 9)
    9

    >>> list_get([1, 2, 3, 4], 2, 8)
    3

    """
    try:
        return l[index]
    except IndexError:
        return default

def get_reference_properties(filename):
    """Figure out ASDF (instrument, filekind, serial) based on `filename`.

    >>> get_reference_properties('tests/data/roman_0001.pmap')
    ('tests/data', 'roman', '', '', '0001', '.pmap')

    >>> get_reference_properties("./roman_wfi_flat.asdf")
    ('.', 'roman', 'wfi', 'flat', '', '.asdf')

    >>> get_reference_properties('tests/data/s7g1700gl_dead_bad_xsum.fits')
    Traceback (most recent call last):
    ...
    crds.core.exceptions.CrdsNamingError: Can't identify instrument of 's7g1700gl_dead_bad_xsum.fits' : Invalid instrument 'cos'

    """
    try:   # Hopefully it's a nice new standard filename, easy
        return decompose_newstyle_name(filename)
    except AssertionError:  # cryptic legacy paths & names, i.e. reality
        pass
    # If not, dig inside the FITS file, slow
    return ref_properties_from_header(filename)

# =======================================================================

def ref_properties_from_header(filename):
    """Look inside ASDF `filename` header to determine instrument, filekind.

    >>> ref_properties_from_header('tests/data/roman_wfi16_f158_flat_small.asdf')
    ('tests/data', 'roman', 'wfi', 'flat', 'roman_wfi16_f158_flat_small', '.asdf')

    >>> ref_properties_from_header('tests/data/s7g1700gl_dead_bad_xsum.fits')
    Traceback (most recent call last):
    ...
    crds.core.exceptions.CrdsNamingError: Can't identify instrument of 's7g1700gl_dead_bad_xsum.fits' : Invalid instrument 'cos'
    """
    # For legacy files,  just use the root filename as the unique id
    path, parts, ext = _get_fields(filename)
    serial = os.path.basename(os.path.splitext(filename)[0])
    header = data_file.get_free_header(filename, (), None, "roman")
    header["ROMAN.META.TELESCOPE"] = "roman"
    name = os.path.basename(filename)
    try:
        instrument = utils.header_to_instrument(header).lower()
        assert instrument in INSTRUMENTS, "Invalid instrument " + repr(instrument)
    except Exception as exc:
        raise exceptions.CrdsNamingError(
            "Can't identify instrument of", repr(name), ":", str(exc)) from exc
    try:
        filekind = header.get('ROMAN.META.REFTYPE', 'UNDEFINED').lower()
        assert filekind in FILEKINDS, "Invalid file type " + repr(filekind)
    except Exception as exc:
        raise exceptions.CrdsNamingError("Can't identify ROMAN.META.REFTYPE of", repr(name))
    return path, "roman", instrument, filekind, serial, ext

# =============================================================================

def reference_keys_to_dataset_keys(rmapping, header):
    """Given a header dictionary for a reference file, map the header back to keys
    relevant to datasets.  So for ACS biasfile the reference says BINAXIS1 but
    the dataset says NUMCOLS.  This would convert { "BINAXIS1": 1024 } to {
    "NUMCOLS" : 1024 }.

    In general,  rmap parkeys are matched against datset values and are defined
    as dataset header keywords.   For refactoring though,  what's initially
    available are reference file keywords...  which need to be mapped into the
    terms rmaps know:  dataset keywords.

    Another aspect of this translation is handling reference file "pattern"
    keywords which typically define or-barred sets of values rather than
    discrete values, any of which the reference is defined to support:
    e.g. 'DETECTOR1|DETECTOR2' vs. 'DETECTOR1'.  In this case, the reference
    file will define a pattern keyword used to define the match pattern in the
    rmap, while a dataset will define a discrete valued keyword which is
    matched on.  e.g. reference file keyword "META.EXPOSURE.P_EXPTYPE" is
    translated back to dataset keyword "META.EXPOSURE.TYPE".  Reference files
    can specify parameters in either form and the P_ pattern variant is given
    preference if both values are defined.  For CRDS purposes, only the P_
    version is checked and used since it will be used to replace the discrete
    valued keyword in the header which is certified and used to define the rmap
    updates.

    Note, can't test unrecognized "P_" keywords because the logging appeares to go to stderr which doctests don't check.

    ==================================================
    Test adding a translation.

    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({'MOUSE' : 'RAT'}, ''), \
    {"MOUSE" : "MICKEY", "RAT" : "MORTIMER"})
    {'MOUSE': 'MICKEY', 'RAT': 'MICKEY', 'ROMAN.META.SUBARRAY.NAME': 'UNDEFINED', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED'}

    ==================================================
    Test replacing translated values with untranslated values.

    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({'MOUSE' : 'RAT'}, ''), \
    {"ROMAN.META.EXPOSURE.P_EXPTYPE" : None, \
    "ROMAN.META.INSTRUMENT.P_BAND" : "UNDEFINED", \
    "ROMAN.META.INSTRUMENT.P_DETECTOR"  : "RADAR", \
    "ROMAN.META.INSTRUMENT.P_CHANNEL" : None, \
    "ROMAN.META.INSTRUMENT.CHANNEL" : None, \
    "ROMAN.META.INSTRUMENT.P_FILTER" : "UNDEFINED", \
    "ROMAN.META.INSTRUMENT.FILTER" : None, \
    "ROMAN.META.INSTRUMENT.P_MODULE" : "LUNAR", \
    "ROMAN.META.INSTRUMENT.MODULE" : None, \
    "ROMAN.META.SUBARRAY.P_SUBARRAY" : None, \
    "ROMAN.META.SUBARRAY.NAME" : "YELLOW", \
    "ROMAN.META.INSTRUMENT.P_GRATING" : "UNDEFINED", \
    "ROMAN.META.INSTRUMENT.GRATING" : "MOZZARELLA", \
    "ROMAN.META.EXPOSURE.PREADPATT" : "CHECKERBOARD", \
    "ROMAN.META.EXPOSURE.READPATT" : "CHESSBOARD"})
    {'ROMAN.META.EXPOSURE.P_EXPTYPE': None, 'ROMAN.META.INSTRUMENT.P_BAND': 'UNDEFINED', 'ROMAN.META.INSTRUMENT.P_DETECTOR': 'RADAR', 'ROMAN.META.INSTRUMENT.P_CHANNEL': None, 'ROMAN.META.INSTRUMENT.CHANNEL': None, 'ROMAN.META.INSTRUMENT.P_FILTER': 'UNDEFINED', 'ROMAN.META.INSTRUMENT.FILTER': None, 'ROMAN.META.INSTRUMENT.P_MODULE': 'LUNAR', 'ROMAN.META.INSTRUMENT.MODULE': 'LUNAR', 'ROMAN.META.SUBARRAY.P_SUBARRAY': None, 'ROMAN.META.SUBARRAY.NAME': 'YELLOW', 'ROMAN.META.INSTRUMENT.P_GRATING': 'UNDEFINED', 'ROMAN.META.INSTRUMENT.GRATING': 'MOZZARELLA', 'ROMAN.META.EXPOSURE.PREADPATT': 'CHECKERBOARD', 'ROMAN.META.EXPOSURE.READPATT': 'CHECKERBOARD', 'ROMAN.META.INSTRUMENT.DETECTOR': 'RADAR', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED'}

    ==================================================
    Test setting missing subarray and exposure type values.

    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({}, ''), \
    {})
    {'ROMAN.META.SUBARRAY.NAME': 'UNDEFINED', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED'}

    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({}, ''), \
    {'ROMAN.META.SUBARRAY.NAME' : 'REDOCTOBER', \
    'ROMAN.META.EXPOSURE.TYPE' : 'NORTHFACE'})
    {'ROMAN.META.SUBARRAY.NAME': 'REDOCTOBER', 'ROMAN.META.EXPOSURE.TYPE': 'NORTHFACE'}

    ==================================================
    Test preserving existing subarry adn exposure type values.

    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({}, ''), \
    {'ROMAN.META.SUBARRAY.NAME' : 'REDOCTOBER', \
    'ROMAN.META.EXPOSURE.TYPE' : 'NORTHFACE'})
    {'ROMAN.META.SUBARRAY.NAME': 'REDOCTOBER', 'ROMAN.META.EXPOSURE.TYPE': 'NORTHFACE'}

    ==================================================
    Test preseverving existing DATE/TIME if no USEAFTER value.

    >>> config.ALLOW_BAD_USEAFTER.reset()
    >>> reference_keys_to_dataset_keys( \
    namedtuple('x', ['reference_to_dataset', 'filename'])({}, 'secret_code_file.txt'), \
    {'ROMAN.META.OBSERVATION.DATE' : '1879-03-14', \
     'ROMAN.META.OBSERVATION.TIME' : '12:34:56'})
    {'ROMAN.META.OBSERVATION.DATE': '1879-03-14', 'ROMAN.META.OBSERVATION.TIME': '12:34:56', 'ROMAN.META.SUBARRAY.NAME': 'UNDEFINED', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED'}

    ==================================================
    Test setting DATE/TIME with no USEAFTER, but allowed "bad use after".

    >>> config.ALLOW_BAD_USEAFTER.reset()
    >>> config.ALLOW_BAD_USEAFTER.set("1")
    False
    >>> reference_keys_to_dataset_keys(namedtuple('x', ['reference_to_dataset', 'filename'])({}, 'secret_code_file.txt'), {})
    {'ROMAN.META.SUBARRAY.NAME': 'UNDEFINED', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED', 'ROMAN.META.OBSERVATION.DATE': '1900-01-01', 'ROMAN.META.OBSERVATION.TIME': '00:00:00'}

    ==================================================
    Test setting DATE/TIME from USEAFTER.

    >>> config.ALLOW_BAD_USEAFTER.reset()
    >>> config.ALLOW_BAD_USEAFTER.set("1")
    False
    >>> reference_keys_to_dataset_keys(namedtuple('x', ['reference_to_dataset', 'filename'])({}, 'secret_code_file.txt'), \
    {'ROMAN.META.USEAFTER' : '1770-12-01T01:23:45', \
     'ROMAN.META.OBSERVATION.DATE' : '1879-03-14', \
     'ROMAN.META.OBSERVATION.TIME' : '12:34:56'})
    {'ROMAN.META.USEAFTER': '1770-12-01T01:23:45', 'ROMAN.META.OBSERVATION.DATE': '1770-12-01', 'ROMAN.META.OBSERVATION.TIME': '01:23:45', 'ROMAN.META.SUBARRAY.NAME': 'UNDEFINED', 'ROMAN.META.EXPOSURE.TYPE': 'UNDEFINED'}

    ==================================================
    Test bad formatted USEAFTER.

    >>> config.ALLOW_BAD_USEAFTER.reset()
    >>> reference_keys_to_dataset_keys(namedtuple('x', ['reference_to_dataset', 'filename'])({}, 'secret_code_file.txt'), \
    {'ROMAN.META.USEAFTER' : 'bad user after', \
     'ROMAN.META.OBSERVATION.DATE' : '1879-03-14', \
     'ROMAN.META.OBSERVATION.TIME' : '12:34:56'})
    Traceback (most recent call last):
    ...
    crds.core.exceptions.InvalidUseAfterFormat: Bad USEAFTER time format = 'bad user after'
    """
    header = dict(header)

    # Basic common pattern translations
    translations = {
            "ROMAN.META.EXPOSURE.P_EXPTYPE" : "ROMAN.META.EXPOSURE.TYPE",

            "ROMAN.META.INSTRUMENT.P_BAND" : "ROMAN.META.INSTRUMENT.BAND",

            "ROMAN.META.INSTRUMENT.P_DETECTOR"  : "ROMAN.META.INSTRUMENT.DETECTOR",

            "ROMAN.META.INSTRUMENT.P_CHANNEL" : "ROMAN.META.INSTRUMENT.CHANNEL",

            "ROMAN.META.INSTRUMENT.P_FILTER" : "ROMAN.META.INSTRUMENT.FILTER",

            "ROMAN.META.INSTRUMENT.P_MODULE"  : "ROMAN.META.INSTRUMENT.MODULE",

            "ROMAN.META.SUBARRAY.P_SUBARRAY" : "ROMAN.META.SUBARRAY.NAME",

            "ROMAN.META.INSTRUMENT.P_GRATING" : "ROMAN.META.INSTRUMENT.GRATING",

            "ROMAN.META.EXPOSURE.PREADPATT" : "ROMAN.META.EXPOSURE.READPATT",
            "ROMAN.META.EXPOSURE.P_READPATT" : "ROMAN.META.EXPOSURE.READPATT",

            # vvvv Speculative,  not currently defined or required by CAL vvvvv
            "ROMAN.META.INSTRUMENT.PCORONAGRAPH" : "ROMAN.META.INSTRUMENT.CORONAGRAPH",
        }

    # Rmap header reference_to_dataset field tranlations,  can override basic!
    try:
        translations.update(rmapping.reference_to_dataset)
    except AttributeError:
        pass

    log.verbose("reference_to_dataset translations:\n", log.PP(translations), verbosity=60)
    log.verbose("reference_to_dataset input header:\n", log.PP(header), verbosity=80)

    for key in header:
        # Match META.X.P_SOMETHING or P_SOMETH
        if (key.split(".")[-1].startswith("P_")) and key not in translations:
            log.warning("CRDS-pattern-like keyword", repr(key),
                        "w/o CRDS translation to corresponding dataset keyword.")
            log.info("Pattern-like keyword", repr(key),
                     "may be misspelled or missing its translation in CRDS.  Pattern will not be used.")
            log.info("The translation for", repr(key),
                     "can be defined in crds.roman.locate or rmap header reference_to_dataset field.")
            log.info("If this is not a pattern keyword, adding a translation to 'not-a-pattern'",
                     "will suppress this warning.")

    # Add replacements for translations *if* the existing untranslated value
    # is poor and the translated value is better defined.   This is to do
    # translations w/o replacing valid/concrete DM values with something
    # like guessed values of "UNDEFINED" or "N/A".
    for rkey in sorted(translations):
        if rkey in header:
            dkey = translations[rkey]
            dval = header.get(translations[rkey], None)
            rval = header[rkey]
            if rval not in [None, "UNDEFINED"] and rval != dval:
                log.info("Setting", repr(dkey), "=", repr(dval),
                         "to value of", repr(rkey), "=", repr(rval))
                header[dkey] = rval

    if "ROMAN.META.SUBARRAY.NAME" not in header:
        header["ROMAN.META.SUBARRAY.NAME"] = "UNDEFINED"

    if "ROMAN.META.EXPOSURE.TYPE" not in header:
        header["ROMAN.META.EXPOSURE.TYPE"] = "UNDEFINED"

    # If USEAFTER is defined,  or we're configured to fake it...
    #   don't invent one if its missing and we're not faking it.
    if "ROMAN.META.USEAFTER" in header or config.ALLOW_BAD_USEAFTER:

        # Identify this as best as possible,
        filename = header.get("ROMAN.META.FILENAME", None) or rmapping.filename

        reformatted = timestamp.reformat_useafter(filename, header).split()
        header["ROMAN.META.OBSERVATION.DATE"] = reformatted[0]
        header["ROMAN.META.OBSERVATION.TIME"] = reformatted[1]

    log.verbose("reference_to_dataset output header:\n", log.PP(header), verbosity=80)

    return header

# =============================================================================

def condition_matching_header(rmapping, header):
    """Normalize header values for .rmap reference insertion.

    >>> condition_matching_header(None, {1:2, 3:4, 5:6})
    {1: 2, 3: 4, 5: 6}

    """
    return dict(header)   # NOOP for Roman,  may have to revisit

# ============================================================================

def fits_to_parkeys(fits_header):
    """
    >>> condition_matching_header(None, {1:2, 3:4, 5:6})
    {1: 2, 3: 4, 5: 6}

    """

    return dict(fits_header)

# ============================================================================

def get_env_prefix(instrument):
    """Return the environment variable prefix (IRAF prefix) for `instrument`.

    >>> get_env_prefix(None)
    'crds://'

    """
    return "crds://"

def filekind_to_keyword(filekind):
    """Return the "keyword" at which a assigned reference should be recorded.

    >>> filekind_to_keyword(None)
    Traceback (most recent call last):
    ...
    NotImplementedError: filekind_to_keyword not implemented for Roman

    """
    raise NotImplementedError("filekind_to_keyword not implemented for Roman")

def locate_file(refname, mode=None):
    """Given a valid reffilename in CDBS or CRDS format,  return a cache path for the file.
    The aspect of this which is complicated is determining instrument and an instrument
    specific sub-directory for it based on the filename alone,  not the file contents.

    The mode=None test case is disabled because it mysteriously causes these tests to
    fail when running the runtests script:
        ERROR: test_throughput_lookup_generation (crds.tests.test_synphot_lookup_generator.TestSynphotLookupGenerator)
        FAIL: Doctest: crds.tests.test_bad_files.dt_bad_references_fast_mode
        FAIL: Doctest: crds.tests.test_bad_files.dt_bad_rules_jwst_getreferences_warning
        FAIL: Doctest: crds.tests.test_certify.certify_recursive
        FAIL: Doctest: crds.tests.test_certify.certify_table_comparison_context
        FAIL: Doctest: crds.tests.test_heavy_client.dt_getreferences_ignore_cache
        FAIL: Doctest: crds.tests.test_list.dt_list_cached_references
        FAIL: Doctest: crds.tests.test_synphot_hst.dt_synphot_core_integration_test
        FAIL: Doctest: crds.tests.test_synphot_hst.dt_synphot_core_integration_test
    XXXX TODO: Enable the mode=None test case and resolve the ensuing test failures in other modules.
    >> locate_file('tests/data/roman_wfi16_f158_flat_small.asdf', None) # doctest: +ELLIPSIS
    '.../references/roman/wfi/roman_wfi16_f158_flat_small.asdf'

    >>> locate_file('tests/data/roman_wfi16_f158_flat_small.asdf', 'instrument') # doctest: +ELLIPSIS
    '.../references/roman/wfi/roman_wfi16_f158_flat_small.asdf'

    >>> locate_file('tests/data/roman_wfi16_f158_flat_small.asdf', 'flat') # doctest: +ELLIPSIS
    '.../references/roman/roman_wfi16_f158_flat_small.asdf'

    >>> locate_file('tests/data/roman_wfi16_f158_flat_small.asdf', 'other') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    ValueError: Unhandled reference file location mode 'other'

    """
    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="roman")
    if mode == "instrument":
        instrument = utils.file_to_instrument(refname)
        rootdir = locate_dir(instrument, mode)
    elif mode == "flat":
        rootdir = config.get_crds_refpath("roman")
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return  os.path.join(rootdir, os.path.basename(refname))

def locate_dir(instrument, mode=None):
    """Locate the instrument specific directory for a reference file.

    The mode=None test case is disabled because it mysteriously causes these tests to
    fail when running the runtests script:
        ERROR: test_throughput_lookup_generation (crds.tests.test_synphot_lookup_generator.TestSynphotLookupGenerator)
        FAIL: Doctest: crds.tests.test_bad_files.dt_bad_references_fast_mode
        FAIL: Doctest: crds.tests.test_bad_files.dt_bad_rules_jwst_getreferences_warning
        FAIL: Doctest: crds.tests.test_certify.certify_recursive
        FAIL: Doctest: crds.tests.test_certify.certify_table_comparison_context
        FAIL: Doctest: crds.tests.test_heavy_client.dt_getreferences_ignore_cache
        FAIL: Doctest: crds.tests.test_list.dt_list_cached_references
        FAIL: Doctest: crds.tests.test_synphot_hst.dt_synphot_core_integration_test
        FAIL: Doctest: crds.tests.test_synphot_hst.dt_synphot_core_integration_test
    XXXX TODO: Enable the mode=None test case and resolve the ensuing test failures in other modules.
    >> locate_dir('wfi', None) # doctest: +ELLIPSIS
    '.../references/roman/wfi'

    >>> locate_dir('wfi', 'instrument') # doctest: +ELLIPSIS
    '.../references/roman/wfi'

    >>> locate_dir('wfi', 'flat') # doctest: +ELLIPSIS
    '.../references/roman'

    >>> locate_dir('wfi', 'other') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    AssertionError: Invalid CRDS cache subdirectory mode = 'other'

    """

    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="roman")
    else:
        config.check_crds_ref_subdir_mode(mode)
    crds_refpath = config.get_crds_refpath("roman")
    if mode == "instrument":   # use simple names inside CRDS cache.
        rootdir = os.path.join(crds_refpath, instrument.lower())
        if not os.path.exists(rootdir):
            if config.writable_cache_or_verbose("Skipping making instrument directory link for", repr(instrument)):
                utils.ensure_dir_exists(rootdir + "/locate_dir.fits")
    elif mode == "flat":    # use original flat cache structure,  all instruments in same directory.
        rootdir = crds_refpath
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return rootdir

def get_cross_strapped_pairs(header):
    """Roman does not use FITS files so there is no cross-strapping between datamodels
    notation and FITS keywords.   Returns []

    >>> get_cross_strapped_pairs(None)
    []

    """
    return []

# ============================================================================

def test():
    """Run the module doctests."""
    import doctest
    from . import locate
    return doctest.testmod(locate)



"""This module defines an abstract API for tables used in CRDS file certification row checks
and bestrefs table effects determinations.  In both cases it basically provides a list of
SimpleTable objects,  one per segment/hdu for a table file and a simple object which gives
readonly row acess to each segment.
"""

import os.path

from astropy import table

from crds.core import utils, log
from crds import data_file

_HERE = os.path.dirname(__file__) or "."

def ntables(filename):
    """Return the number of segments / hdus in `filename`."""
    if filename.endswith(".fits"):
        tables = 0
        with data_file.fits_open(filename) as hdus:
            for i,hdu in enumerate(hdus):
                if "TABLEHDU" in hdu.__class__.__name__.upper():
                    tables += 1
        return tables
    else:
        return 1

@utils.cached
def tables(filename):
    """Return [ SimpleTable(filename, segment), ... ] for each table segment in filename.

    This function is self-cached.    Clear the cache using clear_cache().
    """
    if filename.endswith(".fits"):
        tables = []
        with data_file.fits_open(filename) as hdus:
            for i,hdu in enumerate(hdus):
                classname = hdu.__class__.__name__.upper()
                if "TABLEHDU" in classname:
                    tables.append(SimpleTable(filename, i))
                    if classname == "TABLEHDU":
                        log.warning("ASCII Table detected in HDU#", str(i) +
                                    ".  Particularly for HST, verify that it should not be a BIN Table HDU.")
        return tables
    else:
        return [ SimpleTable(filename, segment=1) ]

def clear_cache():
    """Clear the cached values for the tables interface."""
    tables.cache.clear()


class SimpleTable:
    """A simple class to encapsulate astropy tables for basic CRDS readonly table row and colname access."""
    def __init__(self, filename, segment=1):
        self.filename = filename
        self.segment = segment
        self.basename = os.path.basename(filename)
        self._columns = None  # dynamic,  independent of astropy
        if filename.endswith(".fits"):
            with data_file.fits_open(filename) as hdus:
                tab = hdus[segment].data
                self.colnames = tuple(name.upper() for name in tab.columns.names)
                self.rows = tuple(tuple(row) for row in tab)   # readonly
        else:
            tab = table.Table.read(filename)
            self.colnames = tuple(name.upper() for name in tab.columns)
            self.rows = tuple(tuple(row) for row in tab)   # readonly
        log.verbose("Creating", repr(self), verbosity=60)

    @property
    def columns(self):
        """Based on the row tuples,  create columns dict dynamically.  This permits closing astropy Table during __init__.

        Retuns { colname : column, ... }
        """
        if self._columns is None:
            self._columns = dict(list(zip(self.colnames, list(zip(*self.rows)))))
        return self._columns

    def __repr__(self):
        return (self.__class__.__name__ + "(" + repr(self.basename) + ", " + repr(self.segment) + ", colnames=" +
                repr(self.colnames) + ", nrows=" + str(len(self.rows)) + ")")



def test():
    import doctest, crds.io.tables
    return doctest.testmod(crds.io.tables)

if __name__ == "__main__":
    print(test())



"""This module defines CRDS specific exceptions."""

class CrdsError(Exception):
    """Baseclass for all client exceptions."""
    def __init__(self, *args, **keys):
        super(CrdsError, self).__init__(" ".join(str(arg) for arg in args), **keys)

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(str(self)) + ")"

# -------------------------------------------------------------------------------------------

class ServiceError(CrdsError):
    """The service call failed for some reason."""

class StatusChannelNotFoundError(ServiceError):
    """Requested status channel does not exist.  Typo or deleted."""

class OwningProcessAbortedError(ServiceError):
    """An abort request was recieved on the specified channel."""

# -------------------------------------------------------------------------------------------

class CrdsConfigError(CrdsError):
    """An inconsistency has been detected between CRDS configuration and operating status or data."""

# -------------------------------------------------------------------------------------------

class CrdsWebError(CrdsError):
    """The id='error_message' <div> of a CRDS_server responsae page was present and not empty.
    Generic 'caught' web error with error message.
    """

class CrdsWebAuthenticationError(CrdsWebError):
    """The id='error_login' <div> of a CRDS server login response was present and not empty.
    Login to the website failed due to username or password.
    """
class CrdsExtraneousFileError(CrdsWebError):
    """Unexpected files are already in the ingest directory during a file submisssion."""

# -------------------------------------------------------------------------------------------

class CrdsBackgroundError(CrdsError):
    """An exception occurred in a background task and was re-raised in the foreground."""

# -------------------------------------------------------------------------------------------

class CrdsDownloadError(CrdsError):
    """Error downloading data for a reference or mapping file."""

# -------------------------------------------------------------------------------------------

class CrdsBadRulesError(CrdsError):
    """Attempt to use a bad CRDS rule or context containing a bad rule to determine best references."""

class CrdsBadReferenceError(CrdsError):
    """Attempt to recommend a bad CRDS reference file."""

# -------------------------------------------------------------------------------------------

class CrdsNetworkError(CrdsError):
    """First network service call failed, nominally connection refused."""

class CrdsLookupError(CrdsError, LookupError):
    """Filekind NOT FOUND for some reason defined in the exception string."""

class CrdsRemoteContextError(CrdsError):
    """There was a problem pushing or retrieving the value of a pipeline context
    recorded on the CRDS server, i.e. the pipeline's echo of the context
    they're using as reported by the sync tool.   A remote context is CRDS's
    attempt to track the default context actually sync'ed to a pipeline, i.e.
    what that particular cache says.  In contrast the operational context
    is a global maintained on the CRDS server defining what the default *should
    be* as a result of successful and current cache sync'ing by any pipeline
    or CRDS user.
    """

# -------------------------------------------------------------------------------------------

class FilePropertiesError(CrdsError):
    """Failure to determine instrument/or and reftype."""

class CrdsUnknownInstrumentError(CrdsError):
    """Reference to an instrument which does not exist in a context."""

class CrdsUnknownReftypeError(CrdsError):
    """Reference to a filekind which does not exist in a context."""

class MappingError(CrdsError):
    """Exception in load_rmap."""

class MappingFormatError(MappingError):
    "Something wrong with context or rmap file format."

class ChecksumError(MappingError):
    "There's a problem with the mapping's checksum."

class MissingHeaderKeyError(MappingError):
    """A required key was not in the mapping header."""

class InconsistentParkeyError(MappingError):
    """The parkey tuple was inconsistent with the rest of the .rmap in some way."""

class MappingInsertionError(MappingError):
    """The insertion of a new file into a mapping (rmap) failed for some reason."""

# -------------------------------------------------------------------------------------------

class ValidationError(CrdsError):
    """Some Selector key did not match the set of legal values."""

class AmbiguousMatchError(CrdsError):
    """Represents a MatchSelector which matched more than one equivalently
    weighted choice.   Ambiguous matches represents a problem in the RMAP
    for projects which don't allow these or Selectors which don't support
    them.   NOTE: allowing ambiguous matches is important to HST and the
    canonical HST Match -> UseAfter pattern can support them under many
    circumstances.  The semantics for these are shaky since it's possible
    to have merge collisions which work out badly.
    """

class MissingParameterError(CrdsError):
    """A required parameter for a matching selector did not appear
    in the parameter dictionary.
    """

class BadValueError(CrdsError):
    """A required parameter for a matching selector did not have
    any of the valid values.
    """

class ModificationError(CrdsError):
    """Failed attempt to modify rmap, e.g. replacement vs. addition."""

class OverlappingMatchError(ModificationError):
    """At least two of the references being added to an rmap had exactly the
    same matching criteria (for at least one of their applicable cases) causing
    one of the two submissions to be omitted from the update.
    """

class MatchingError(CrdsLookupError):
    """Represents a MatchSelector lookup which failed."""

class UseAfterError(CrdsLookupError):
    """None of the dates in the selector precedes the processing date."""

class VersionAfterError(CrdsLookupError):
    """Look up using version failed."""

# -------------------------------------------------------------------------------------------

class InvalidDatetimeError(CrdsError):
    """CRDS was unable to parse the value from date/time related keywords."""

class InvalidUseAfterFormat(InvalidDatetimeError):
    """CRDS was unable to parse the value from a USEAFTER keyword or equivalent."""

class InvalidVersionFormat(CrdsError):
    """CRDS was unable to parse the value from a CAL_VER keyword or equivalent."""

# -------------------------------------------------------------------------------------------

class IrrelevantReferenceTypeError(LookupError):
    """The reference determined by this rmap does not apply to the instrument
    mode specified by the dataset header.

    Based on the "rmap_relevance" rmap header expression.
    """

class OmitReferenceTypeError(LookupError):
    """The reference determined by this rmap does not apply to the instrument
    mode specified by the dataset header,  and should be completely omitted from
    the bestrefs results dictionary.

    Based on the "rmap_omit" rmap header expression.
    """

# -------------------------------------------------------------------------------------------

class FileFormatError(CrdsError):
    """Something was seriously wrong with a file's format."""

class JsonFormatError(FileFormatError):
    """What should be valid JSON didn't parse / load."""

class YamlFormatError(FileFormatError):
    """What should be valid YAML didn't parse / load."""

class UnsupportedFileOpError(CrdsError, NotImplementedError):
    """In CRDS,  some function is not supported for a particular file format."""

# -------------------------------------------------------------------------------------------

class CertifyError(CrdsError):
    """Certify exception baseclass."""

class MissingKeywordError(CertifyError):
    """A required keyword was not defined."""

class IllegalKeywordError(CertifyError):
    """A keyword which should not be defined was present."""

class InvalidFormatError(CertifyError):
    """The given file was not load-able."""

class TpnDefinitionError(CertifyError):
    """Something was wrong with a .tpn constraint."""

class MissingTpnIncludeError(CertifyError):
    """Any .tpn include files must exist or it's an error.  This is in contrast to other speculative .tpn loads."""

class InconsistentTpnReplaceError(CertifyError):
    """Within the closure of a .tpn text stream,  the same value was mentioned by "replace" twice... with different replacements."""

class TypeSetupError(CertifyError):
    """An error occured while trying to locate file constraints and row mode variables."""

class MissingReferenceError(CertifyError):
    """A reference file mentioned by a mapping isn't in CRDS yet."""

class RequiredConditionError(CertifyError):
    """An ExpressionValidator evaluated to False with respect to a file header."""

class BadKernelCenterPixelTooSmall(CertifyError):
    """A convolution kernel array had a center pixel <= 1.0."""

class BadKernelSumError(CertifyError):
    """A convolution kernel array had a sum too far from 1.0, or similar."""

class MissingArrayError(CrdsError):
    """A specified (named?) array was not found in a reference file."""

class MissingColumnError(CrdsError):
    """A specified (named?) column was not found in a reference file."""

class DuplicateSha1sumError(CrdsError):
    """This file is bit-for-bit identical with a file already in CRDS."""

class CrdsTypeSpecError(CrdsError):
    """Something was wrong with a CRDS type spec registering a type with CRDS."""

# -------------------------------------------------------------------------------------------

class NameComparisonError(CrdsError):
    """Failed to determine the time order of two names."""

class CrdsNamingError(CrdsError):
    """CRDS enountered an error(s) while trying to name a file."""

# -------------------------------------------------------------------------------------------

class CrdsPipelineCfgDeterminationError(CrdsError):
    """Problem determining cal code pipeline .cfg for a dataset."""

class CrdsPipelineTypeDeterminationError(CrdsError):
    """The process of inferring reference types from cal code pipeline .cfg files failed."""

# -------------------------------------------------------------------------------------------

_exception_names = [ name for name in dir() if name.endswith("Error") ]

__all__ = _exception_names



"""Module makes submit package run as a program using crds.submit"""
import sys

from . import submit

def main():
    sys.exit(submit.ReferenceSubmissionScript()())

if __name__ == "__main__":
    main()



"""Fixed September 2013 hooks for ACS (biasfile)

Replaced NUMROWS, NUMCOLS with NAXIS2, NAXIS1 everywhere.

Restored simple runtime fallback_header approach.

Added "hooks" to newly generated rmap headers provided acs_biasfile_filter is imported.

Simplified rmap generation process.
"""
import copy

from .acs_common import *

# ===========================================================================

def precondition_header_acs_biasfile_v2(rmap, header_in):
    """Mutate the incoming dataset header based upon hard coded rules
    and the header's contents.   This is an alternative to generating
    an equivalent and bulkier rmap.
    """
    header = dict(header_in)
    log.verbose("acs_biasfile_precondition_header:", log.format_parameter_list(header))
    exptime = timestamp.reformat_date(header["DATE-OBS"] + " " + header["TIME-OBS"])
    if (exptime < SM4):
        #if "APERTURE" not in header or header["APERTURE"] == "UNDEFINED":
        log.verbose("Mapping pre-SM4 APERTURE to N/A")
        header["APERTURE"] = "N/A"
    try:
        naxis1 = int(float(header["NAXIS1"]))
    except ValueError:
        log.verbose("acs_biasfile_selection: bad NAXIS1.")
        # sys.exc_clear()
    else:
        header["NAXIS1"] = utils.condition_value(str(naxis1))
        # if pre-SM4 and NAXIS1 > HALF_CHIP
        exptime = timestamp.reformat_date(header["DATE-OBS"] + " " + header["TIME-OBS"])
        if (exptime < SM4):
            if naxis1 > ACS_HALF_CHIP_COLS:
                if header["CCDAMP"] in ["A","D"]:
                    log.verbose("acs_bias_file_selection: exposure is pre-SM4, converting amp A or D " +
                                "to AD for NAXIS1 = " + header["NAXIS1"])
                    header["CCDAMP"] = "AD"
                elif header["CCDAMP"] in ["B","C"]:
                    log.verbose("acs_bias_file_selection: exposure is pre-SM4, converting amp B or C " +
                                "to BC for NAXIS1 = " + header["NAXIS1"])
                    header["CCDAMP"] = "BC"
    if header['DETECTOR'] == "WFC" and \
        header['XCORNER'] == "0.0" and header['YCORNER'] == "0.0":
        log.verbose("acs_biasfile_selection: precondition_header halving NAXIS2")
        try:
            naxis2 = int(float(header["NAXIS2"])) / 2
        except ValueError:
            log.verbose("acs_biasfile_selection: bad NAXIS2.")
            # sys.exc_clear()
        else:
            header["NAXIS2"] = utils.condition_value(str(naxis2))
    dump_mutations(header_in, header)
    return header

def dump_mutations(header1, header2):
    log.verbose("In header1 not header2:", set(dict(header1).items()) - set(dict(header2).items()))
    log.verbose("In header2 not header1:", set(dict(header2).items()) - set(dict(header1).items()))

# ===========================================================================

#   This section contains matching customizations.

BIASFILE_PARKEYS = ('DETECTOR', 'CCDAMP', 'CCDGAIN', 'APERTURE', 'NAXIS1', 'NAXIS2', 'LTV1', 'LTV2', 'XCORNER', 'YCORNER', 'CCDCHIP')
# , ('DATE-OBS', 'TIME-OBS')),

def fallback_header_acs_biasfile_v2(rmap, header_in):
    """Mutates dataset header for 2nd try when primary match fails."""
    header = precondition_header_acs_biasfile_v2(rmap, header_in)
    log.verbose("No matching BIAS file found for",
               "NAXIS1=" + repr(header['NAXIS1']),
               "NAXIS2=" + repr(header['NAXIS2']),
               "LTV1=" + repr(header['LTV1']),
               "LTV2=" + repr(header['LTV2']))
    log.verbose("Trying full-frame default search")
    if header['DETECTOR'] == "WFC":
        header["NAXIS1"] = "4144.0"
        header["NAXIS2"] = "2068.0"
        header["LTV1"] = "24.0"
        header["LTV2"] = "0.0"
    else:
        header["NAXIS1"] = "1062.0"
        header["NAXIS2"] = "1044.0"
        header["LTV1"] = "19.0"
        if header['CCDAMP'] in ["C","D"]:
            header["LTV2"] = "0.0"
        else: # assuming HRC with CCDAMP = A or B
            header["LTV2"] = "20.0"
    return header

def acs_biasfile_filter(kmap_orig):
    """
    Post-SM4 APERTURE's of '' were all replaced and cannot match,  hence dropped at CRDS rmap generation time.
    """
    kmap = copy.deepcopy(kmap_orig)
    header_additions = {
        "hooks" : {
            "precondition_header" : "precondition_header_acs_biasfile_v2",
            "fallback_header" : "fallback_header_acs_biasfile_v2",
            },
        }
    for match in kmap_orig:
        header = dict(list(zip(BIASFILE_PARKEYS, match)))
        if header["APERTURE"] == "":
            for filemap in kmap_orig[match]:
                if filemap.date > SM4:
                    log.warning("Removing empty post-SM4 APERTURE for", repr(filemap))
                else:  # conceptually,  N/A.   To prevent conincidental stronger matches: *.
                    log.warning("Setting empty pre-SM4 APERTURE to 'N/A' for", repr(filemap))
                    new_match = match[0:3] + ("N/A",) + match[4:]
                    if new_match not in kmap:
                        kmap[new_match] = []
                    if filemap not in kmap[new_match]:
                        kmap[new_match].append(filemap)
                        kmap[new_match].sort()
                kmap[match].remove(filemap)
            if not kmap[match]:
                log.warning("Removing empty match", repr(match))
                del kmap[match]
    return kmap, header_additions

def acs_darkfile_filter(kmap_orig):
    """
    Post-SM4 APERTURE's of '' were all replaced and cannot match,  hence dropped at CRDS rmap generation time.
    """
    kmap = copy.deepcopy(kmap_orig)
    header_additions = {}
    for match in kmap_orig:
        header = dict(list(zip(BIASFILE_PARKEYS, match)))
        try:
            if float(header["CCDGAIN"]) == -999.0:
                log.warning("CCDGAIN=-999.0 Deleting match", match, "with", kmap[match])
                del kmap[match]
        except Exception:
            pass
    return kmap, header_additions
